from dataclasses import dataclass, field
from typing import List, Dict, Any
import pandas as pd
import cobra
import itertools
import re
import json 
import numpy as np

from scarcc.util import (convert_arg_to_list, remove_Zero_col)
from scarcc.preparation.metabolic_model import get_gene_id
# from growth_summary import (get_maximum_growth_cycle)

def get_rcts_list(model, gcomb_list): 
    rcts_list = list()
    rcts_set = set()
    for i, gene in enumerate(gcomb_list):
        gene_rcts = [rct.id for rct in model.genes.get_by_id(get_gene_id(model, gene)).reactions]
        if i > 0:
            gene_rcts = list(set(gene_rcts) - rcts_set)
        rcts_list.append(gene_rcts)
        rcts_set = rcts_set | set(gene_rcts)
    return rcts_list

def adjust_flux_df(model, df, gene_combo: list, alpha_table:pd.DataFrame): # model for query of reactions
    # Don't use v1 cols to indicate gene_inhibition, will skip unidirectional reaction
    print('time x', model, gene_combo)
    def query_alpha(gene_combo, alpha_table):
        splitted = gene_combo.split('.')
        print(splitted)
        if len(splitted) <= 2: # not checkerboard 
            gcomb_alpha = {gene: alpha_table.loc[gene, f'{model.id}'] for gene in gene_combo.split('.')}
            return gcomb_alpha, alpha_table

        # alphas
        else:
            splitted = gene_combo.split('_')
            # Gene_inhibition, lv_pair = [ele.split('.') for ele in 'folP.folA_1.2'.split('_')]
            Gene_inhibition, lv_pair = [ele.split('.') for ele in gene_combo.split('_')]
            lv_pair = tuple([int(ele) for ele in lv_pair])
            gcomb_alpha = dict()
            alpha_table = alpha_table.query('lv_pairs == @lv_pair')
            for current_gene in Gene_inhibition:
                # print(alpha_table.head())
                gcomb_alpha.update({current_gene:
                                    alpha_table.loc[current_gene, f'{model.id}']})
            return gcomb_alpha, alpha_table

    print(gene_combo)

    if 'Normal' not in gene_combo:
#         gene_combo_dict = get_gcomb_alpha_dict(gene_combo) 
        v1_cols = df.filter(regex='v1').columns
        
        # gcomb_alpha = {gene: alpha_table.loc[gene, f'{model.id}'] for gene in gene_combo.split('.')}
        gcomb_alpha, alpha_table = query_alpha(gene_combo, alpha_table)
        # print(gcomb_alpha)
        gcomb_alpha = dict(sorted(gcomb_alpha.items(), key=lambda item: item[1], reverse=True))
        
        rcts_list = get_rcts_list(model, gcomb_alpha.keys()) # exclude repeated rct for two gene
        scaled_rcts = list()
    # return df
        for gene, rcts in zip(gcomb_alpha.keys(), rcts_list):
#             rcts = [rct for rct in rcts if rct in orig_cols]
            for orig_col in rcts:
                alpha = alpha_table.loc[f'{gene}', f'{model.id}']
                v1_col = orig_col + "_v1"
                reversible = v1_col  in v1_cols
                if reversible: 
                    df[f'{orig_col}'] = (df[f'{orig_col}'] + df[f'{v1_col}'])/alpha # only forward or backward != 0  
                    df = df.drop(f'{v1_col}', axis=1)  
                else:
                    df[f'{orig_col}'] = (df[f'{orig_col}'])/alpha # only forward or backward != 0              
    return df

def get_XG_cycle_from(desired_cycle):
    if len(desired_cycle.index[-1].split('.')) <=2:
        SG_cycle = desired_cycle.loc[[len(ele.split('.')) ==1 for ele in desired_cycle.index]]
        DG_cycle = desired_cycle.loc[[len(ele.split('.')) ==2 for ele in desired_cycle.index]]
    else:
        SG_cycle = desired_cycle.loc[['0' in ele for ele in desired_cycle.index]]
        DG_cycle = desired_cycle.loc[['0' not in ele for ele in desired_cycle.index]]
    SG_cycle.index.name='SG'
    DG_cycle.index.name='DG'
    SG_cycle.columns.name=None
    DG_cycle.columns.name=None
    return SG_cycle, DG_cycle    


def get_growth_phase_summary(model, culture, df_from_js, desired_cycle, current_gene_combo, is_flux=True):
    
    growth_phase = desired_cycle.query('Species == @model.id & culture == @culture').loc[current_gene_combo, 'growth_phase']
    growth_df = df_from_js.query('cycle>@growth_phase[0] & cycle<@growth_phase[1]').drop('cycle', axis=1)
#     growth_df = growth_df.drop(['cycle','x','y'], axis=1) if is_flux else growth_df.drop(['cycle','x','y'])
    summary_df = (growth_df.apply(['mean','std'])
         .unstack().to_frame().T)
    summary_df.columns = summary_df.columns.map('_'.join).str.strip('_')
    summary_df.index= [current_gene_combo]
    return summary_df


def set_GI_SP_as_MI(end_BM, manual_SI = False, multi_index = None):
    if type(end_BM.index[0]) is not int: # Gene_inhibition is not index # Series.dtype -> dtype('int64')
        if end_BM.index.name in [None, 'DG']:
            end_BM.index.name = 'Gene_inhibition'
        if end_BM.index.name == 'Gene_inhibition':
            if 'Gene_inhibition' in end_BM.columns:
                end_BM = end_BM.drop('Gene_inhibition', axis=1)
            end_BM = end_BM.reset_index()
    if multi_index:
        return end_BM.set_index(multi_index)
    if 'Species' in end_BM.columns and manual_SI == False:
        return end_BM.set_index(['Gene_inhibition', 'Species'])
    else:
        return end_BM.set_index('Gene_inhibition')


def join_dfs_using_MI(df_list, how='left', multi_index = None):
    df_list = [set_GI_SP_as_MI(df, multi_index=multi_index) for df in df_list]
    result_df = df_list[0]
    for df in df_list[1:]:
        result_df = result_df.join(df, how=how)
    return result_df

all_culture_items = ['coculture_media', 'E0_coculture_flux', 'S0_coculture_flux', 'E0_monoculture_media', 'E0_monoculture_flux', 'S0_monoculture_media', 'S0_monoculture_flux']
def get_desired_keys(culture_options, item_options, all_culture_items=all_culture_items):
    culture_options, item_options = '|'.join(culture_options), '|'.join(item_options)
    desired_keys = [ele for ele in all_culture_items if re.search(culture_options, ele) and re.search(item_options, ele)]
    return desired_keys


def retrive_specific_culture(file_list, culture_options, item_options, XG_df_list):
    out_dict = dict()
    # # mono_dict = dict()
    # file_list = j
    # file_list, XG_df_list = convert_arg_to_list(file_list), [SG_cycle, DG_cycle]
    first_file = file_list[0]
    for file, XG_df in itertools.zip_longest(file_list, XG_df_list, fillvalue=first_file):
        if isinstance(file, str):
            file = json.load(open(file))
        XG_list = XG_df.index

        for outer_key in get_desired_keys(culture_options=culture_options, item_options=item_options):
            out_dict['_'.join([outer_key, XG_df.index.name])] = {current_gene: file[outer_key][current_gene] for current_gene in XG_list}                                            
    return out_dict

def get_alpha_wide(alpha_table):
    return alpha_table.melt(value_vars=['E0', 'S0'], var_name='Species', value_name='alpha', ignore_index=False)

# def get_SG_DG(DG_list, explode=True):
#     df = (pd.DataFrame(pd.DataFrame(2*[DG_list]
#     , index=['Gene_inhibition','SG']).T
#     ).set_index('Gene_inhibition'))
#     return df.explode('SG') if explode else df

def get_SG_DG(DG_list, explode=True):
    df = pd.DataFrame(pd.DataFrame(2*[DG_list], index=['Gene_inhibition','SG']).T.set_index('Gene_inhibition').SG.str.split('.'))
    return df.explode('SG') if explode else df

def get_alpha_to_merge(alpha_table, DG_list): # TODO: check necessity- use in normal drug comb
    # def get_alpha(l: list[str, str], alpha_table):
    #     l=l[0]
    #     df = pd.DataFrame(alpha_table.loc[l, ['E0', 'S0']]).T
    #     return df.values.tolist()
    def get_alpha(l: list[str, str], alpha_table):
        if isinstance(l, pd.Series):
            l=l.to_list()
        l=l[0]
        df = pd.DataFrame(alpha_table.loc[l, ['E0', 'S0']]).T
        return df.values.tolist()


    # S = [ele for ele in alpha_table.columns if 'S0' in ele]
    ES_cols = ['E0', 'S0']
    alpha_wide = get_alpha_wide(alpha_table)
    df = (get_SG_DG(DG_list, explode=False)
          .apply(lambda x: pd.Series(get_alpha(x, alpha_table), index=ES_cols), axis=1) # df with E0 S0 alpha columns
          .melt(value_vars=ES_cols, var_name='Species', value_name='alpha', ignore_index=False)
          .combine_first(alpha_wide))
    return df

def modify_checkerboard_flux_compare_df(alpha_table, df):
    def get_ICX_from_lv_pair_list(alpha_table, lv_pair_list=[4.2, 5.3]):
        ICX = dict()
        corrected_lv_pair_list = list()
        print(lv_pair_list)
        for lv_pair in convert_arg_to_list(lv_pair_list):
            if isinstance(lv_pair, str):
                lv_pair = str(lv_pair).split('.')
                lv_pair = tuple([int(x) for x in lv_pair])
            corrected_lv_pair_list.append(lv_pair)
            # print(alpha_table.query('lv_pairs==@lv_pair').ICX.unique())
            ICX.update({lv_pair: alpha_table.query('lv_pairs==@lv_pair').ICX.to_dict()})
        return ICX 

    def get_lv_pairs_col(flux_compare_df):
        lv_pairs = [lv_pair[1].split('.') for lv_pair in flux_compare_df.index.str.split('_')]
        lv_pairs = [tuple([int(ele) for ele in lv_pair]) for lv_pair in lv_pairs]
        return lv_pairs
    df['lv_pairs'] = get_lv_pairs_col(df)
    cols = df.columns.tolist()
    df['ICX'] = df.apply(lambda x : get_ICX_from_lv_pair_list(alpha_table, [x.lv_pairs]).values(), axis=1)
    cols = ['Species', 'XG', 'lv_pairs', 'ICX'] + [col for col in cols if col not in ['Species', 'XG', 'lv_pairs']]
    df = df[cols]
    return df

@dataclass(kw_only=True)
class FluxCompare:
    desired_cycle : pd.DataFrame
    file_list : None
    E0 : cobra.Model = None
    S0 : cobra.Model = None
    # species_options : List = field(default_factory=lambda: [E0, S0])
    culture_options : str = field(default_factory=lambda: ['coculture'])
    item_options : str = field(default_factory=lambda: ['media', 'flux'])
    flux_dict : Dict = field(default_factory=dict)
    flux_compare_dict : Dict = field(default_factory=dict)
    flux_compare_df : pd.DataFrame = field(default_factory=pd.DataFrame)
    desired_cycle_col : str = 'cycle_max_gr'
    is_checkerboard :bool = False
    alpha_table :pd.DataFrame = None
    count = 0
    DG_list : List = None

    def __post_init__(self):
        self.SG_cycle, self.DG_cycle = get_XG_cycle_from(self.desired_cycle)
        self.SG_list, self.DG_list = set(self.SG_cycle.index), set(self.DG_cycle.index)
        self.XG_df_list = [self.SG_cycle, self.DG_cycle]
        self.culture_options, self.item_options = convert_arg_to_list(self.culture_options), convert_arg_to_list(self.item_options)
        self.file_list = convert_arg_to_list(self.file_list)
        if 'lv_pairs' in self.alpha_table.columns:
            self.is_checkerboard = True

    def retrive_specific_culture(self):
        out_dict = dict()
        first_file = self.file_list[0]
        for file, XG_df in itertools.zip_longest(self.file_list, self.XG_df_list, fillvalue=None):
            if isinstance(file, str):
                print('load from json file', str)
                file = json.load(open(file))
                previous_js = file
            elif file is None:
                file = previous_js

            XG_list = XG_df.index

            for outer_key in get_desired_keys(culture_options=self.culture_options, item_options=self.item_options):
                out_dict['_'.join([outer_key, XG_df.index.name])] = {current_gene: file[outer_key][current_gene] for current_gene in XG_list}                                            
        self.flux_dict = out_dict
        return None
    
    def get_flux_dict_per_culture_item(self, model, sub_flux_dict, culture_item, desired_cycle, log_step=globals().get('log_step', 5), high_pass_cycle=450): # desired_cycle as SG_cycle, DG_cycle
        def get_flux_snapshot(s : pd.Series, no_grow, temp_flux_dict, culture, cg):
            
            cycle=s[self.desired_cycle_col]
            current_gene_combo = s.name
            print(model.id, culture_item, self.count)
            if cycle < 10:
                no_grow.append(current_gene_combo)
                temp_flux_dict[current_gene_combo] = pd.DataFrame(index = [f'{current_gene_combo}'])
            else:
                desired_js = sub_flux_dict.get(f'{current_gene_combo}')
                # self.s = desired_js
                # coculture_dict['E0_monoculture_flux_SG']['folP.folA_1.0'] 

                df_from_js = pd.DataFrame.from_dict(desired_js)
                is_flux = '_flux' in culture_item

                if is_flux:
                    df_from_js = adjust_flux_df(model, df_from_js, current_gene_combo,alpha_table=self.alpha_table, is_checkerboard=self.is_checkerboard).drop(['x','y'], axis=1)
                        
                else:
                    df_from_js = df_from_js.pivot(index=['cycle'],columns=['metabolite'], values='conc_mmol').reset_index()
                growth_phase_summary = get_growth_phase_summary(model, culture, df_from_js, desired_cycle, current_gene_combo, is_flux=is_flux)
                

                invalid_snapshot, max_correction = True, 3
                while invalid_snapshot:

                    invalid_snapshot=False
                    print(cg ,'cyc', cycle, 'mxc', max_correction, max_correction <= -1)
                    temp_df = df_from_js.query("cycle == @cycle")
 
                    temp_df.index = [current_gene_combo]
                    temp_df = pd.concat([temp_df, growth_phase_summary], axis=1) # mean, sd for each reaction -> ncols*3
                    print('snap&m',float(abs(temp_df['EX_bulk_ac_e'])),float(abs(temp_df['EX_bulk_ac_e_mean'])))
                    if model.id == 'S0' or cycle>high_pass_cycle or np.isnan(temp_df['EX_bulk_ac_e_mean'][0]) or float(abs(temp_df['EX_bulk_ac_e'])) > float(abs(temp_df['EX_bulk_ac_e_mean']))*.7:
                        invalid_snapshot = False # stop iter
                    else:
                        max_correction -= 1
                        cycle -= log_step                        
                        if max_correction <= -1 or cycle < 10:
                            invalid_snapshot = False # stop iter
                            temp_df.loc[current_gene_combo, 'EX_bulk_ac_e'] = temp_df.loc[current_gene_combo, 'EX_bulk_ac_e_mean']

                temp_flux_dict.update({current_gene_combo: temp_df}) 
                print('kkk', temp_flux_dict.keys())
            return temp_flux_dict, no_grow
            
        temp_flux_dict, no_grow = dict(), list()
        culture = culture_item.split('_')[0]
        for current_gene_combo, cycle_row in (desired_cycle
                                              .query('Species == @model.id & culture == @culture')
                                            #   .iloc[:1]
                                              .iterrows()):
            print(current_gene_combo,'currr')
            temp_flux_dict, no_grow = get_flux_snapshot(cycle_row, no_grow, temp_flux_dict, culture, current_gene_combo)
            
        # print(temp_flux_dict)
        if no_grow:
            print('Zero growth: ', ', '.join(no_grow))
        if not temp_flux_dict:
            return pd.DataFrame()
        # print('vvvvvvvvvvvv---------', len(temp_flux_dict.values()))

        return pd.concat(temp_flux_dict.values()).copy()

    def get_flux_compare_df(self): # generation of mono_flux, coculture_flux, exclude media
        temp_flux_dict = dict()
        if not self.flux_dict:
            self.flux_dict = self.retrive_specific_culture()

        for key, sub_flux_dict in self.flux_dict.items(): 
            print(key)
            if 'media' not in key:
                current_species, current_XG, culture = key.split('_')[0], key.split('_')[-1], key.split('_')[1]
                XG_cycle = self.SG_cycle if current_XG == 'SG' else self.DG_cycle
                model = self.E0 if current_species == 'E0' else self.S0
                current_species = model.id
                culture_item = culture + '_flux' 

                print(current_species, current_XG)
                if key not in temp_flux_dict:
                    flux_df = self.get_flux_dict_per_culture_item(model, sub_flux_dict, culture_item, XG_cycle) # flux snapshot for all GI as index 
                
                flux_df['Species'] = current_species
                flux_df['XG'] = current_XG
                flux_df['culture'] = culture
                temp_flux_dict[key] = flux_df # flux_df all gene_inhibition as index
                print(key, '--------', flux_df.index)
                
        flux_compare_df = pd.concat(temp_flux_dict.values()).copy()
        flux_compare_df.index.name = 'Gene_inhibition'
        flux_compare_df = remove_Zero_col(flux_compare_df)
    #     flux_compare_df = get_ESdiff(remove_Zero_col(flux_compare_df)) # ES diff 41 columns ?only common
    #     flux_compare_df = get_Ndiff(flux_compare_df)
        self.flux_compare_df = flux_compare_df
        

        # if not self.is_checkerboard: #  alpha value not joined for checkerboard, SG format different 
        #     flux_compare_df = join_dfs_using_MI([get_alpha_to_merge(self.alpha_table, self.DG_list), flux_compare_df]
        #                                         , how='right') # DG only
        # else:
        #     flux_compare_df = modify_checkerboard_flux_compare_df(self.alpha_table, flux_compare_df)
        # self.flux_compare_df = flux_compare_df
        return flux_compare_df
    




from dataclasses import dataclass
import pandas as pd
from typing import List
from abc import ABC, abstractmethod
from typing import Union
import logging

logger = logging.getLogger(__name__)

@dataclass(kw_only=True)
class AlphaFinderConfig(ABC):
    """Configure the alpha finder class with procedures update next alpha and classification of gene

    Parameters
    ----------
    current_gene : str
        gene name
    trace_obj_val : dict
        objective value of the gene
    is_growth_switch : bool
        whether the gene is growth switch, biomass drop abruptly within certain precision of alpha, mostly transport reaction in energy pathway
    early_stop : bool
        reached maximum iteration - without reaching the target precision of alpha
    ko_intercepted : bool
        stop finding alpha since the gene is non-essential, producing no effect to change in biomass
    is_monoculture : bool
        whether the alpha search is done in FBA, monoculture
    model : Union["cobra.Model", List["cobra.Model"]]
        single species in FBA
        list of species in COMETS coculture simulation 
    ever_eval : bool
        initialize first run in coculture
    target_obj_val : float
        target biomass value produced by the perturbation by alpha
    search_alpha : float
        alpha value to be started with, updated in each iteration
    alpha_lb : float
        lower bound of alpha, objective value is higher than target one, updated in each iteration
    alpha_ub : float
        upper bound of alpha, objective value is lower than target one, updated in each iteration
    precision : int
        precision of alpha
    is_new_ub : bool
        whether the upper bound is updated in the current iteration
    exp_leap : float
        multiplier factor of search_alpha in exponential search if new_ub is False, implement binary search otherwise
    found_alpha : bool
        whether the alpha in inhibiting biomass yield to target biomass is found
    iter_max : int
        maximum iteration for alpha search
    i_iter : int
        current iteration of alpha search, initialized with 0 
    acceptance_threshold_upper : float
        multiplier of acceptance threshold for target biomass, value higher than 1
    acceptance_threshold_lower : float
        lower bound of acceptance threshold for target biomass, value between 0 and 1

    Remarks
    -------
    Configuration of alpha finder class for MonocultureAlphaFinder and CocultureAlphaFinder, DO NOT use this class directly
    """

    # gene characteristic and records
    current_gene: str = None
    trace_obj_val: dict = None
    is_growth_switch: bool = None
    early_stop: bool = False
    ko_intercepted: bool = False

    # culture identification
    is_monoculture: bool = None
    model: Union["cobra.Model", List["cobra.Model"]] = None # single model for monoculture, list for coculture
    ever_eval: bool =  False # initialize first run coculture

    # alpha search related
    target_obj_val : float = None
    search_alpha: float = None
    alpha_lb : float = 1+1e-7
    alpha_ub : float = 1e5
    precision : int = 2 # precision of alpha
    is_new_ub : bool = None
    exp_leap : float = 2
    found_alpha : bool = None
    iter_max: int = 25
    i_iter: int = 0

    # evaluation alpha choice
    acceptance_threshold_upper : float = None
    acceptance_threshold_lower : float = None

    @abstractmethod
    def eval_alpha_fun(self):
        pass

    @abstractmethod
    def out_fun(self):
        pass

    @staticmethod
    def get_next_alpha(search_alpha, alpha_lb, alpha_ub, is_new_ub, exp_leap=2):
        """Update alpha search value in each iteration if target biomass is not reached"""
        if is_new_ub is False:  # raise lb, search higher alpha
            alpha_lb = search_alpha
            if (search_alpha*exp_leap <alpha_ub and is_new_ub is False):
                search_alpha *= exp_leap
            else: # search alpha not get updated, then binary search
                search_alpha = (search_alpha+alpha_ub)/2
        else: # search alpha not accepted, lower ub, lower search_alpha
            alpha_ub = search_alpha
            search_alpha = max((search_alpha+alpha_lb)/2, 1+1e-5) # terminate at 1+1e-5
        return search_alpha, alpha_lb, alpha_ub
    
    @staticmethod
    def is_found(search_alpha, alpha_lb, alpha_ub, precision):
        """Check whether the alpha is found within the precision"""
        if (search_alpha<2) and (precision <3):
            precision = 3
        if alpha_lb > 15:
            precision = 1
        return (round(search_alpha,precision)==round(alpha_ub,precision) or
                round(search_alpha,precision)==round(alpha_lb,precision) or
                (search_alpha>9e4))

    def classify_growth_switch(self, min_attain=.9, max_attain=.3) -> bool:
        """Classify the gene as growth switch
        biomass drop abruptly within certain precision of alpha, mostly transport reaction in energy pathway
        need to reach min_attain and max_attain for biomass value to classify as growth switch, 
        risk of misclassification otherwise
        Precision adjusted for alpha range

        Parameters
        ----------
        min_attain : float
            minimum biomass value to be attained
        max_attain : float
            maximum biomass value to be attained

        Returns
        -------
        bool
            whether the gene is growth switch
        """
        alpha_range = (self.alpha_ub - self.alpha_lb)
        alpha_range_narrow = (((alpha_range < 0.15) and (self.alpha_ub < 3)) or # circumvent precision <2 in subsequent evaluation
                                ((alpha_range < .3) and (5<self.alpha_lb<10)) or
                                ((alpha_range < 1.3) and (self.alpha_lb > 10)))
        alpha_range_req =  (alpha_range_narrow
                            and (self.alpha_lb > 1.01)) # ever update lb and ub
        
        obj_req = (not any(0.3 < val < 0.8 for val in self.trace_obj_val) and # mrdA forever > 1 for low dose
                    (min(self.trace_obj_val) < min_attain) and
                    (max(self.trace_obj_val) > max_attain))
        
        if alpha_range_narrow and all(val >1 for val in self.trace_obj_val): # small dose all > Normal
            obj_req = True
                
        self.is_growth_switch = (alpha_range_req and obj_req) or (self.alpha_ub < 1.018)
        return self.is_growth_switch
    
    def find_feasible_alpha(self):
        """If non-essential, producing no effect to change in biomass, jump to output function
        If essential, find the alpha that can inhibit the biomass yield to target biomass

        Returns
        -------
        output of out_fun defined in child class
        """
        logger.debug(self.current_gene, self.ko_intercepted)
        def eval_continue():
            # see if conditions for stopping are met
            if self.is_monoculture: # as culture_flag
                return not self.found_alpha
            if self.i_iter>self.iter_max:
                self.early_stop = True
                logger.debug('Early stopped coculture')
            return ((not self.found_alpha) or (self.i_iter<2)
                    or (self.alpha_lb < 1.01 and 1.018< self.alpha_ub)) # force iter -ensure optimal

        if self.ko_intercepted: # req run ko_gr first, otherwise cannot catch
            logger.debug(f'Intercept Non-essential: {self.current_gene}, calculating with current alpha')
            return self.out_fun() # sim_culture with current alpha_table already ran
                
        if not self.ever_eval:
            self.eval_alpha_fun() # initial ko is without alpha, only used for identify gr_ko

        stop = not (self.alpha_lb < self.search_alpha < self.alpha_ub)
        
        while eval_continue() and not stop: 
            self.search_alpha, self.alpha_lb, self.alpha_ub = self.get_next_alpha(
                self.search_alpha, self.alpha_lb, self.alpha_ub, self.is_new_ub,
                self.exp_leap)
            self.eval_alpha_fun()
            self.i_iter+=1
        if (not self.is_monoculture):
            logger.debug(f'Stopped at iter {self.i_iter}') if (self.i_iter<=self.iter_max or self.i_iter ==2) else logger.debug('Success search, end at: ', str(self.i_iter))
        
        return self.out_fun()



import os
import logging
import pandas as pd
from typing import List
from dataclasses import dataclass, field

import concurrent.futures
import cometspy as c

from scarcc.util import convert_arg_to_list
from .simulation_configuration import LayoutConfig
from scarcc.preparation.perturbation import get_alphas_from_tab, alter_Sij
from .result_processing import extract_biomass_flux_df

logger = logging.getLogger(__name__)

def sim_culture(layout, p=None, base = None):
    # flux data needs biomass to determine the ~, only return sim object for the outpout object
    # separate function into mono & coculture to prevent using wrong layer
    if isinstance(layout, list):
        if len(layout) > 1:
            raise ValueError("The list 'layout' should contain only one element.")
        (layout,) = layout # one element unpacking from iter_species
    
    sim = c.comets(layout, p)
    sim.working_dir = os.path.join(base, '') # make sure it is a directory instead of file
    print(sim.working_dir)
    
    try:
        sim.run()
    except:
        logging.exception(f"{sim.run_output}")
        print(f"{sim.run_output}")
    return sim

@dataclass(kw_only=True)
class CombinedAntibioticsSimulation(LayoutConfig):
    current_gene: str
    p: 'comets.p'
    alpha_table: str
    base: str = None # base as __file__ or working directory
    checker_suffix: str = None
    return_sim: bool = False
    ko: bool = False

    # default values for output
    working_dir: str = None # passed to comets object
    biomass_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    co_sim_object: 'comets.simulation' = field(default_factory=list)
    mono_sim_object_list: List['comets.simulation'] = field(default_factory=list)
    sim_object_list: List['comets.simulation'] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.current_gene = convert_arg_to_list(self.current_gene)

        path_elements = [self.base, 'SimChamber', '.'.join(self.current_gene)] if 'SimChamber' not in self.base else [self.base, '.'.join(self.current_gene)]
        self.working_dir = os.path.join(*path_elements) # '' as specification of directory where COMETS files are stored
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)
        # filepath
    def cleanup(self):
        try:
            os.rmdir(self.working_dir)
        except:
            print(f'Failed to remove {self.working_dir}, needs remove files manually')

    def get_BM_df(self):
        with self.E0 as m_E0, self.S0 as m_S0:
            metabolic_model_list = [m_E0, m_S0]
            if 'Normal' not in self.current_gene:
                alphas = [get_alphas_from_tab(model, genes=self.current_gene, alpha_table=self.alpha_table) for model in metabolic_model_list]
                _ = [alter_Sij(model, alphas=alpha, genes=self.current_gene, ko=self.ko) for model, alpha in zip(metabolic_model_list, alphas)]

            E_model, S_model = self.set_comets_model()
            co_layout, E0_layout, S0_layout = self.set_layout_object()

            if self.co:
                logger.debug(f'{self.p.all_params["maxCycles"]} co_p')
                self.co_sim_object = sim_culture(self.co_layout, p=self.p, base=self.working_dir)
                self.sim_object_list.append(self.co_sim_object)

            if self.mono:
                monoculture_to_run = dict()
                if self.mono_E:
                    monoculture_to_run[self.E_model] = self.monoE_layout
                if self.mono_S:
                    monoculture_to_run[self.S_model] = self.monoS_layout
                self.mono_sim_object_list = [sim_culture(layout, p=self.p, base=self.working_dir) for layout in monoculture_to_run.values()]
                self.sim_object_list.extend(self.mono_sim_object_list)
            
        self.biomass_df, self.flux_df = extract_biomass_flux_df(self.E0, self.S0, self.sim_object_list, alpha_table=self.alpha_table, current_gene=self.current_gene)
        self.biomass_flux_dict = {'biomass': self.biomass_df, 'flux': self.flux_df}
        self.cleanup()
        return self.biomass_df, self.flux_df

# TODO: add multiprocessing for parsing gene combo list 


import os
def find_directory(desired_directory_name, script_filepath): # start with where the script is located
    """Find directory in the current directory or up maximum 2 levels of directory. 
    If not found, make directory at script location.
    
    Parameters
    ----------
    
    desired_directory_name : str
        Name of the directory to find or make.

    script_filepath : str
        Path to the script that is calling this function.

    Returns
    -------
    directory_path : str

    """

    # def check_directory(directory_path):
    #     return os.path.isdir(directory_path)
    
    # start_directory = os.path.abspath(script_filepath)
    # current_directory = start_directory

    # max_depth = 2
    # while os.path.basename(current_directory) != 'scarccpy' and max_depth > 0:
    #     # Move up one level
    #     current_directory = os.path.dirname(current_directory)
    #     max_depth -= 1

    #     directory_path = os.path.join(current_directory, desired_directory_name)
    #     if check_directory(directory_path):
    #         return directory_path

    # directory_path = os.path.join(os.path.dirname(start_directory), desired_directory_name)
    # print(f'Directory {desired_directory_name} not exist, make directory at script level: {directory_path}')
    # os.makedirs(directory_path)
    # return directory_path

    desired_directory = os.path.join(script_filepath, desired_directory_name)
    if not os.path.isdir(desired_directory):
        print('make directory at script level')
        os.makedirs(desired_directory)
    return desired_directory


"""This module contains functions for retriving components of a model"""
import re
import cobra
from typing import Union, List, Dict
from scarcc.util import convert_arg_to_list

def get_gene_id(model: "cobra.Model", gene_name: str) -> str:
    """Query the id of the gene
    
    Parameters
    ----------
    model: cobra.Model
    gene_name: str
    """
    for i in model.genes:
        if i.name == gene_name:
            return i.id

def get_all_components(E0, S0) -> Dict[str, Dict[str, List]]:
    """Returns all the components of the model
    
    Parameters
    ----------
    E0: cobra.Model
    S0: cobra.Model

    Returns
    -------
    Structured Dictionary:
        all_components = {'metabolites': {'E0': [met1, met2, ...], 'S0': [met1, met2, ...]},
                        'genes': {'E0': [gene1, gene2, ...], 'S0': [gene1, gene2, ...]}, 
                        'reactions': {'E0': [rxn1, rxn2, ...], 'S0': [rxn1, rxn2, ...}}
        }
    """

    all_metabolites = {
        model.id: model.metabolites for model in [E0, S0]
    }
    all_genes = {
        model.id: 
            [gene.name for gene in model.genes if not isinstance(gene, str)]
            for model in [E0, S0]
    }

    all_reactions = {
        model.id: model.reactions for model in [E0, S0]
    }

    all_components = {'metabolites': all_metabolites,
                    'genes': all_genes, 
                    'reactions': all_reactions}
    return all_components

def substrate_only(rxn: "Reaction", met: Union["Metabolite" , str]) -> bool:
    """Returns True if the metabolite is a substrate but not product in the reaction
    
    Parameters
    ----------
    rxn: cobra.Reaction
    met: cobra.Metabolite
    
    Returns
    -------
    bool
    """
    return rxn.lower_bound>=0 and met in rxn.reactants  

def substrate_any(rxn: "Reaction", met: Union["Metabolite" , str]):
    """Returns True if the metabolite can be a substrate in the reaction"""
    return (rxn.lower_bound>=0 and met in rxn.reactants) or rxn.lower_bound<0

def product_only(rxn: "Reaction", met: Union["Metabolite" , str]):
    """Returns True if the metabolite is a product but not substrate in the reaction
    Excludes secretion fluxes"""
    if re.search(r"_e$", rxn.id): # secretion flux
        return True
    return (rxn.lower_bound>=0 and met in rxn.products) or (rxn.upper_bound<=0 and met in rxn.reactants)   
    
def get_component(model, query, all_components):
    """Returns the component of the model that matches the query"""
    ID = model.id.split('.')[0]
    if query in all_components['genes'][ID]:
        return model.genes.get_by_id(get_gene_id(model, query))
    if query in all_components['metabolites'][ID]:
        return model.metabolites.get_by_id(query)
    return model.reactions.get_by_id(query)
    
def get_links_component(model: "Model", query: Union["Metabolite" , "Reaction", "Gene", str], all_components: Dict, id_only: bool = True, 
                        is_sub_only: bool = None, is_any_sub: bool = None, is_any_prod: bool = None, is_prod_only: bool = None):
    """Returns the components linked to the query
    
    Parameters
    ----------
    model: cobra.Model
        The model to search

    query: str or list
        The component to search for, can be a list of genes, metabolites or reactions
    
    all_components: dict
        The dictionary containing all the components of the model
    
    id_only: bool
        If True, returns the id of the components
    
    is_sub_only: bool
        If True, returns only the reactions where the query is a substrate
    
    is_any_sub: bool
        If True, returns the reactions where the query is a substrate
    
    is_any_prod: bool
        If True, returns the reactions where the query is a product
    
    is_prod_only: bool
        If True, returns only the reactions where the query is a product
        
    Returns
    -------
    list
        A list of the components linked to the query
        [component1, component2, ...] or [component1.id, component2.id, ...]
    """

    def get_result(model, query, all_components):        
        # query for single component
        query = get_component(model, query, all_components)
        if isinstance(query, cobra.Reaction):
            result = set(query.metabolites)
        else:
            result = query.reactions
            if isinstance(query, cobra.Metabolite):
                if is_sub_only is not None:
                    result = set([rxn for rxn in result if substrate_only(rxn, query)==is_sub_only])
                elif is_any_sub is not None or is_any_prod is not None:
                    result = set([rxn for rxn in result if substrate_any(rxn, query)==is_any_sub])
                elif is_prod_only is not None:
                    result = set([rxn for rxn in result if product_only(rxn, query)==is_prod_only])
        return result
    
    result=set()
    for ele in convert_arg_to_list(query):
        result.update(get_result(model, ele, all_components))
    if id_only:
            result = set([ele.id for ele in result])
    return list(result)



"""iterate items to pass into function
Can be used to iterate models and the corresponding alpha values
extra arguments can be passed to the function
""" 

def iter_species(models,f,*args,**kwargs) -> list: 
    """Factory for passing extra_objects iterate with model into functions.
    
    Parameters
    ----------
    models : list or zip
        List of models or zip of models and extra_objects that will iterate together.
    f : function
        Function to be called with in put as model and extra_objects.
    *args : positional arguments following models or models and extra_objects
    **kwargs : keyword arguments

    Returns
    -------
    r_object : list
        List of return objects from function f.

    Examples
    --------
    # print model names
    >>> _ = iter_species([E0, S0], print)
    E0
    S0 
    # print model names and corresponding alpha as extra_objects
    >>> _ = iter_species(zip([E0, S0],[20,10]), print)
    E0 [20]
    S0 [10]
    """
    def simple_iter():
        for model in models:
            return_list.append(f(model,*args,**kwargs))

    # extra_object iterative with models
    def coupled_iter():
        for model, *extra_objects in models:
            return_list.append(f(model,extra_objects,*args,**kwargs))

    return_list = []
    iter_fun = coupled_iter if isinstance(models, zip) else simple_iter
    iter_fun()
    return(return_list)



__author__ = 'Harcombr Lab'
__version__ = '0.0.1'

# from scarcc import (sim_preparation,
#                     perturb,
#                     sim_engine,
#                     data_analysis)
# from scarcc import sim_preparation
# from scarcc.util import util


import logging
logger = logging.getLogger(__name__)

from . import get_links_component, get_component

def rename_ac_in_bulk(model, id):
    
    new_id = f'bulk_{id}'
    model.metabolites.get_by_id(id).id = new_id
    if 'ac' in id:
        model.metabolites.get_by_id(new_id).name = 'bulk Acetate'
        model.metabolites.get_by_id(new_id).formula = 'C20H30O20'
    if '_e' in id:
        model.reactions.get_by_id(f'EX_{id}').id = f'EX_{new_id}'

def weight_carbon_byproduct(E0, S0, all_components, ac_scale=None, gal_scale=None):
    if gal_scale is not None:
        query_gal = ['gal_p','gal_c']
        if gal_scale > 1:
            gal_scale = -1*(1-1/gal_scale) 
        
        rxns_to_scale = get_links_component(E0, query_gal, all_components, id_only=False, is_prod_only=True) # ['GALt2pp', 'GAL1PPpp', 'GALabcpp', 'GALS3', 'LACZpp', 'GALM2pp', 'LACZ'] are scaled
        logger.debug('reactions in gal scaled: %s', rxns_to_scale)
        for rxn in rxns_to_scale:
            # TODO: only scale LACZpp is desirable, make gal secretion similar to knockout
            metab_to_scale = rxn.metabolites
            rxn.add_metabolites({k:v*gal_scale for k,v in metab_to_scale.items()})
    # E0.reactions.GALtex.knock_out()
    
    if ac_scale is not None: # flux do not count in minflux
        add_scale = ac_scale-1 if ac_scale>=1 else ac_scale
        
        # 0.1 ac_p + 10 h_p <=> 10 ac_c + 10 h_c
        for rxn in ['ACt2rpp']: 
            if isinstance(rxn, str):
                rxn = get_component(E0, rxn, all_components)
            if not 'EX_' in rxn.id:
                metab_to_scale = rxn.metabolites
                rxn.add_metabolites({k:v*add_scale for k,v in metab_to_scale.items()}) 



import re
from dataclasses import dataclass
from typing import Dict, Union, List
import pandas as pd

from scarcc.data_analysis.flux.flux_snapshot import adjust_flux_df
from scarcc.data_analysis import get_desired_cycle
from scarcc.util import convert_arg_to_list

@dataclass
class SimObjectBase:
    # checkerboard handled in adjust_flux_df automatically
    E0: 'cobra.Model'
    S0: 'cobra.Model'
    sim_object: 'comets.solution'
    alpha_table: pd.DataFrame
    current_gene: str

    def __post_init__(self):
        if isinstance(self.current_gene, list):
            self.current_gene = '.'.join(self.current_gene)
        self.biomass_df = (self.sim_object.total_biomass.set_index('cycle'))
        self.model_dict = {'E0': self.E0, 'S0': self.S0}
        self.species_list = [self.model_dict[ele] for ele in self.biomass_df.columns]
        self.culture = 'coculture' if len(self.species_list)>1 else f'monoculture'
        self.biomass_df = self.biomass_df.add_suffix(f'_{self.current_gene}_{self.culture}') # ?need to use rename_columns?

def get_flux_snapshot(sob: SimObjectBase, model: 'cobra.Model' = None): # to map
    model_id = model.id
    print('snnnn', model_id)
    biomass_df = sob.biomass_df.filter(regex=f'{model_id}') # suffix already added
    flux_df = sob.sim_object.fluxes_by_species[f'{model_id}'].copy()
    flux_df['Species'] = model_id
    flux_df['culture'] = sob.culture
    flux_df['Gene_inhibition'] = sob.current_gene
    flux_df = adjust_flux_df(model, flux_df, sob.current_gene, alpha_table=sob.alpha_table)
    
    # get the last cycle
    desired_cycle = get_desired_cycle(biomass_df)
    # return desired_cycle
    # TODO: snapshot handler
    snap_shot_cycle = int(desired_cycle.cycle_max_gr)
    snap_shot = flux_df.query('cycle == @snap_shot_cycle')
    return snap_shot

def extract_biomass_flux_df(E0: 'cobra.Model', S0: 'cobra.Model', sim_object_list: List['comets.solution'], alpha_table: pd.DataFrame, current_gene: str):
    # function for construct class and extraction
    def extract_biomass_flux_df_per_sim(sob: SimObjectBase) -> Dict[str, Union[pd.DataFrame, List]]: # list concat all together altogether for efficiency
        # this is the output to the get_BM_df function
        biomass_flux_dict = dict()
        biomass_flux_dict['biomass'] = sob.biomass_df
        # biomass_flux_dict['flux'] = pd.concat([get_flux_snapshot(sob, model_id=ele) for ele in sob.species_list])
        biomass_flux_dict['flux'] = [get_flux_snapshot(sob, model=ele) for ele in sob.species_list]
        return biomass_flux_dict['biomass'], biomass_flux_dict['flux']
    
    def construct_sob(sim_object: 'comets.solution'):
        return SimObjectBase(E0=E0, S0=S0, sim_object=sim_object, alpha_table=alpha_table, current_gene=current_gene)
    
    # zip(biomass_df_list, flux_df_list)
    return zip(*[extract_biomass_flux_df_per_sim(construct_sob(sim_object)) for sim_object in convert_arg_to_list(sim_object_list)])

def rename_columns(df):
    df.columns = [re.sub('S0_ac_','S0.ac_', ele) for ele in df] # S0_ac -> S0.ac
    df.columns = [re.sub('S0_gal_','S0.gal_', ele) for ele in df] # S0_ac -> S0.ac
    df.columns = [re.sub(',','.',
           re.sub('\'|\(|\)| |\[|\]','',ele)) # ('gene1', 'gene2') -> gene1.gene2
           for ele in df.columns]
    return(df.columns)
    
def gene_index_culture_col_df(analysis_df): 
    analysis_df['Gene_inhibition'] =  ['.'.join(map(str, convert_arg_to_list(l))) for l in analysis_df.Gene_inhibition] # SG, DG, checkerboard g1.g2 form
    analysis_df = analysis_df.set_index('Gene_inhibition')
    return analysis_df



"""functions for stoichiometry scaling"""

import pandas as pd
from typing import List
from scarcc.util import convert_arg_to_list
from scarcc.preparation.metabolic_model import get_gene_id

def scale_reaction(model: "Model", reaction_id: str, alpha: float, direction='forward'):
    """
    scale a reaction by alpha, change the direction of the reaction if needed
    
    Parameters
    ----------
    
    model: cobra.Model
    reaction_id: str
    alpha: float
    direction: str
        'forward': scale product in the reaction 
        'backward': scale reactant in the reaction
    """
    if direction == 'forward':
        model.reactions.get_by_id(reaction_id).lower_bound = 0
        dict_product = dict((k, v) for k, v in model.reactions.get_by_id(reaction_id).metabolites.items() if v >= 0)    # obtain 'end product'
    else:  # reverse reaction
        model.reactions.get_by_id(reaction_id).upper_bound = 0
        dict_product = dict((k, v) for k, v in model.reactions.get_by_id(reaction_id).metabolites.items() if v <= 0)
    if isinstance(alpha, pd.Series):
        alpha = float(alpha)

    for product, unit in dict_product.items():  # scale corresponding metabolite units involved in the reaction
        model.reactions.get_by_id(reaction_id).add_metabolites({product: -unit*(1-1/alpha)}) # only change unit of unidirection metabolite

def separate_reaction(model: "Model", reaction_id: str, alpha: float) -> List:
    """
    decompose reversible reaction into forward and backward
    scale each end product by 1/alpha

    Parameters
    ----------
    model: cobra.Model
    reaction_id: str
    alpha: float
    
    Returns
    -------
    list of reaction ids that are regulated by the target genes
    """
    (lb, ub) = model.reactions.get_by_id(reaction_id).bounds 
    rxn_ids = [reaction_id] #? 
    if(lb < 0 and ub !=0): # only perform if reaction is bidirectional
        rev_reaction = model.reactions.get_by_id(reaction_id).copy() # copy of target reaction 
        rev_reaction.id = f'{reaction_id}_v1' # redefine id for copy of reaction
        model.add_reactions([rev_reaction]) # add to model
    
        scale_reaction(model, rev_reaction.id, alpha, direction='backward')      
        
        rxn_ids.append(rev_reaction.id) # list of id for reaction and reaction_v1 
    scale_reaction(model, reaction_id, alpha, direction='forward')
    return(rxn_ids)

def alter_Sij(model: "Model", alphas: float = 1, genes: str = 'folA', ko=False):
    """Change stoichiometry of reactions encoded by target genes to alpha * original stoichiometry
    
    Parameters
    ----------
    model: cobra.Model
    alphas: float or list of float
        scaling factor for target genes
    genes: str or list of str
        gene id or gene name
    ko: bool
        knock out the reaction if True, otherwise scale stoichiometry
    
    Returns
    -------
    list of reaction ids that are regulated by the target genes
    """
    # get objective value for corresponding alpha
    alphas= convert_arg_to_list(alphas[0]) if isinstance(alphas, list) and len(alphas)==1 else convert_arg_to_list(alphas)  # unlist one layer from zip comprehension 
    genes = convert_arg_to_list(genes)
    
    genes_dict = {gene: alpha for gene, alpha in zip(genes, alphas)}
    genes_sorted = sorted(genes_dict.items(), key=lambda x:x[1], reverse=True) #sort by magnitude of alpha
    rxn_ids = list() # store list of id of reaction and reaction_v1 that regulated by the same gene 
    for current_gene, alpha in genes_sorted:
        current_gene = current_gene.split('_')[0] # for step_alpha
        for rxn in model.genes.get_by_id(get_gene_id(model, current_gene)).reactions:
            if ko:
                model.reactions.get_by_id(rxn.id).knock_out()
            elif (rxn.id not in rxn_ids):
                rxn_ids.extend(separate_reaction(model, rxn.id, alpha))# copy of reaction, forward_terminal_change = True
    return(rxn_ids)

def get_alphas_from_tab(model: "Model", genes: List, alpha_table: pd.DataFrame) -> List: # the returned df of one gcomb index only pass as series
    """Get alpha values from alpha_table for target genes

    Parameters
    ----------
    model: cobra.Model
    genes: list
        gene id or gene name
        alpha_table: pd.DataFrame
            alpha values for genes and reactions
    
    Returns
    -------
    list of alpha values for target genes
    """
    if isinstance(alpha_table, pd.Series):
        alpha_table = alpha_table.to_frame().T
    genes = convert_arg_to_list(genes)
    alphas = [alpha_table.loc[gene ,model.id] for gene in genes]
    return alphas


# import
# import scarcc
# from importlib import reload
# import scarcc.preparation.find_directory
# reload(scarcc.preparation.find_directory)
# from scarcc.preparation.find_directory import find_directory

# data_directory = find_directory('models', __file__)
# print('out', data_directory)

from scarcc.data_analysis.flux.carbon_allocation import get_carbon_allocation_summary
from scarcc.data_analysis.drug_combination_response.classification import convert_po_col
# from scarcc.data_analysis import convert_po_col


from collections.abc import Iterable
from typing import Union, List

def initialize_medium():
    """Returns a dictionary of default medium conditions for the model.
    Medium is without carbon source.
    10 as default value for COMETS maximum uptake rate"""
    nutrient_medium = {'EX_ca2_e': 10, 
                        'EX_cl_e': 10,
                        'EX_cobalt2_e': 10,
                        'EX_cu2_e': 10,
                        'EX_fe2_e': 10,
                        'EX_fe3_e': 10,
                        'EX_k_e': 10,
                        'EX_mg2_e': 10,
                        'EX_mn2_e': 10,
                        'EX_mobd_e': 10,
                        'EX_ni2_e': 10,
                        'EX_o2_e': 10,
                        'EX_pi_e': 10,
                        'EX_so4_e': 10,
                        'EX_zn2_e': 10,
                        'EX_nh4_e': 10}
    return nutrient_medium

def change_medium(model: "cobra.Model", metabolite: Union["Metabolite" , List['Metabolite']], value: float, return_medium=False): # function for setting medium metabolite value 
    """Changes the value of a metabolite/ list of metabolite in the medium
    
    Parameters
    ----------
    model: cobra.Model
    metabolite: cobra.Metabolite
    value: float
    
    Returns
    -------
    None or dict
    
    """
    medium = model.medium
    if not isinstance(value, Iterable):
        metabolite, value = [metabolite], [value]
    for m, v in zip(metabolite, value):
        medium[m] = v
    model.medium = medium
    if return_medium:
        return model.medium



from collections import defaultdict
import pandas as pd
import itertools
import numpy as np

from scarcc.util import convert_arg_to_list
from scarcc.preparation.metabolic_model import get_gene_id

if 'alpha_table' not in globals():
    print('set alpha table None')
    alpha_table = None

label_df = pd.read_excel(open('./Data/iML1515_GP.xlsx', 'rb'), # ? S0 rct pathway label necessary?
              sheet_name='iML1515_GP', index_col=0)

def record_rct(model, current_gene):
    return [rct.id for rct in model.genes.get_by_id(get_gene_id(model, current_gene)).reactions]

def pwy_dict_to_df(pwy_dict):
    df = pd.DataFrame.from_dict({k: ' + '.join(v) for k, v in pwy_dict.items()},
#     df = pd.DataFrame.from_dict({k: '+'.join(v) for k, v in pwy_dict.items()},
                       orient='index', columns=['Pathway'])
    df.index.name = 'Gene_inhibition'
    return df

def convert_dict_to_long_df(d: dict, cols: list):
    gcomb_pwy_list = list()
    for gcomb, pwys in d.items():
        for pwy in pwys:
            gcomb_pwy_list.append([gcomb, pwy])
#     index_col = [col for col in cols if col in 'Gene_inhibition|Reaction']
    df = pd.DataFrame.from_records(gcomb_pwy_list, columns=cols)
    return df

def get_SG_pwy_dict(model, alpha_table=alpha_table):
    gene_rct_dict = defaultdict(list)
    for current_gene in alpha_table.index:
        gene_rct_dict[f'{current_gene}'] = record_rct(model, current_gene)
    
    gene_pwy_dict = defaultdict(list)
    for current_gene,rcts in gene_rct_dict.items():
        for current_reaction in rcts:
            try:
                label = label_df.loc[f'{current_reaction}', 'm_subsystem']
                label = label if type(label) == str else label[0]
                gene_pwy_dict[f'{current_gene}'].append(label)
            except:
                print(current_gene, f'reaction {current_reaction} not exist in list')
        gene_pwy_dict[f'{current_gene}'] = set(convert_arg_to_list(gene_pwy_dict[f'{current_gene}'])) | set(gene_pwy_dict[f'{current_gene}'])
    return gene_pwy_dict

def get_gene_pathway_df(model, alpha_table=alpha_table):
    pwy_dict = get_SG_pwy_dict(model, alpha_table=alpha_table)
    DG_pwy_dict = get_DG_pwy_dict(pwy_dict)
    
    pwy_dict.update(DG_pwy_dict)
    
    gene_pathway_df = convert_dict_to_long_df(pwy_dict, ['Gene_inhibition', 'Pathway'])
    gene_pathway_df['XG'] = gene_pathway_df.Gene_inhibition.str.split('.').apply(
        lambda x: 'SG' if len(x)==1 else 'DG') 
    return gene_pathway_df

def antagonistic_count(genes:list):
    antagonistic_set = ['dadX.rffG', 'dadX.pyrD', 'acnB.thrB', 'acnB.thrB', 'acnB.thrB',
       'aroA.argD', 'aroA.argD', 'aroA.argD', 'dadX.guaB', 'aroA.dapB',
       'aroA.guaB']
    return list((set(genes).intersection(antagonistic_set)))

def get_ratio_col(gene_count_df):
    DG = gene_count_df.query('XG=="DG"').reset_index(drop=True).copy()
    DG['Antagonistic_gcomb'] = DG.Gene_list.apply(lambda x: antagonistic_count(x.split(', ')))
#     DG['Antagonistic_ratio'] = DG.Antagonistic_gcomb.str.split(', ').map(len)
    DG['Antagonistic_ratio'] = pd.Series(map(len, DG.Antagonistic_gcomb.tolist()))/DG.Gene_count
    DG['Antagonistic_gcomb'] = DG['Antagonistic_gcomb'].apply(lambda x: ', '.join(x))
    
    return DG[['Pathway', 'Antagonistic_ratio', 'Antagonistic_gcomb']]

def get_gene_count_df(gene_pathway_df):

    gene_count_df = (gene_pathway_df.groupby(['Pathway','XG'], as_index=False).agg(
                        Gene_list=('Gene_inhibition', lambda x: ', '.join(x.values)),
                        Gene_count=('Gene_inhibition', 'count')))
    gene_count_df = gene_count_df.merge(get_ratio_col(gene_count_df), on='Pathway')
    (gene_count_df.pivot(index=['Pathway','Antagonistic_ratio', 'Antagonistic_gcomb'], columns=['XG'], values=['Gene_list','Gene_count'])
         .sort_values(by=('Gene_count','DG'), ascending=False)).to_csv('./Data/gene_count.csv')
    return gene_count_df

def get_single_pathway_df(gene_pathway_df):
    # add p_o afterwards
    df = (gene_pathway_df.groupby(['Gene_inhibition','XG'], as_index=False).agg(
        Pathway_count=('Pathway', 'count'),
        Pathway_list = ('Pathway', lambda x: list(x.values))))
#         P = ('Pathway', lambda x: pd.DataFrame(x.values))))
    df['Single_pathway'] = pd.cut(df['Pathway_count'], [0,1,10], labels = ['Single_pathway', 'Multi_pathway'])
    return df.sort_values(['Pathway_count', 'XG'], ascending=True).set_index('Gene_inhibition')

def get_rct_pathway_df():
    pwy_rct_dict = defaultdict(list)
    for index, sub_df in label_df[['m_subsystem']].iterrows():
        pwy_rct_dict[sub_df[0]].append(index)
    rct_pathway_df = convert_dict_to_long_df(pwy_rct_dict,['Pathway','Reaction'])
    return rct_pathway_df.drop_duplicates()

def get_DG_pwy_dict(SG_pwy_dict):
    comb_list = list(pd.read_csv('./Data/gr_Div_DG_Blis_Aug31.csv').gene_inhibition[1:])
    DG_pwy_dict = dict()
    # for DG_pwy in comb_list:
    gene_pathway_row = list()
    for DG_pwy in comb_list:
        gene_pair = DG_pwy.split('.')
        DG_pwy_dict[DG_pwy] = SG_pwy_dict[gene_pair[0]] | SG_pwy_dict[gene_pair[1]] 
    return DG_pwy_dict

# reaction column
def get_SG_pwy_df(model, E0, S0, alpha_table):
    def remove_nan(x: pd.Series): #=function remove_nan_from
        return x.apply(lambda x: sorted(list(itertools.compress(x,[ele not in [None, np.nan] for ele in x]))))

    SG_pwy_df = pd.DataFrame.from_dict(get_SG_pwy_dict(model, alpha_table=alpha_table), orient='index')
    SG_pwy_df['Pathways'] = SG_pwy_df.values.tolist()
    SG_pwy_df = pd.DataFrame(remove_nan(SG_pwy_df.Pathways))
#     SG_pwy_df = pd.DataFrame(SG_pwy_df.Pathways
#                              .apply(lambda x: remove_nan(x)))
    SG_pwy_df['Reactions_E0'] = list(pd.Series(SG_pwy_df.index).apply(lambda x: record_rct(E0,x)))
    SG_pwy_df['Reactions_S0'] = list(pd.Series(SG_pwy_df.index).apply(lambda x: record_rct(S0,x)))    
    SG_pwy_df.index.name='Gene_inhibition'
    return SG_pwy_df

def find_reactions_not_exist_in(model, reactions: list):
    if len(reactions) ==0:
        return reactions
    reactions = convert_arg_to_list(reactions)
    reactions_all = [rct.id for rct in model.reactions]
    not_found_reactions = [rct for rct in reactions if rct not in reactions_all]
    return not_found_reactions

def get_pwy_rxn_df(model, E0, S0, alpha_table, DG_list):
    def get_pwys_from(SG_pwy_df, SG):
        return SG_pwy_df.loc[SG, 'Pathways']
    SG_pwy_df = get_SG_pwy_df(model, E0, S0, alpha_table)
    g_pwy_rct_df = pd.DataFrame(DG_list, columns=['Gene_inhibition'])
    g_pwy_rct_df['SG_list'] =  g_pwy_rct_df.apply(lambda x: x.Gene_inhibition.split('.'), axis=1)
    g_pwy_rct_df['Pathways'] = g_pwy_rct_df.apply(lambda x: list(set(SG_pwy_df.loc[x.SG_list[0],'Pathways']) | set(SG_pwy_df.loc[x.SG_list[1],'Pathways'])), axis=1)
    g_pwy_rct_df['Common_pathway'] = g_pwy_rct_df.apply(lambda x: set(get_pwys_from(SG_pwy_df, x.SG_list[0]))
                                                        .intersection(set(get_pwys_from(SG_pwy_df, x.SG_list[1])))
                                                        ,axis=1)
    g_pwy_rct_df['Reactions_E0'] = g_pwy_rct_df.apply(
        lambda x: list(set(SG_pwy_df.loc[x.SG_list[0],'Reactions_E0']) | set(SG_pwy_df.loc[x.SG_list[1],'Reactions_E0'])), axis=1)
    g_pwy_rct_df['Reactions_S0'] = g_pwy_rct_df.apply(
        lambda x: list(set(SG_pwy_df.loc[x.SG_list[0],'Reactions_S0']) | set(SG_pwy_df.loc[x.SG_list[1],'Reactions_S0'])), axis=1)
    g_pwy_rct_df['Common_Reactions'] = g_pwy_rct_df.apply(lambda x: set(x['Reactions_E0']).intersection(set(x['Reactions_S0'])), axis=1)
    g_pwy_rct_df['Reactions_E0_only'] = g_pwy_rct_df.apply(lambda x: set(x['Reactions_E0'])-set(x['Reactions_S0']), axis=1)
    g_pwy_rct_df['Reactions_S0_only'] = g_pwy_rct_df.apply(lambda x: set(x['Reactions_S0'])-set(x['Reactions_E0']), axis=1)
    g_pwy_rct_df['Reactions_not_in_S0'] = g_pwy_rct_df.apply(lambda x: find_reactions_not_exist_in(S0, x.Reactions_E0_only), axis=1)
    g_pwy_rct_df['Reactions_not_in_E0'] = g_pwy_rct_df.apply(lambda x: find_reactions_not_exist_in(E0, x.Reactions_S0_only), axis=1)
    g_pwy_rct_df = pd.concat([g_pwy_rct_df.set_index('Gene_inhibition'), SG_pwy_df])
    return g_pwy_rct_df




import pandas as pd

def convert_po_col(col, additive_threshold=0.05):
    diff_bins = [-10,-1*additive_threshold,1*additive_threshold,10] 
    return pd.cut(col, bins=diff_bins, labels=['Antagonistic', 'Additive', 'Synergistic'])



# from flux_ratio_analysis import (get_p_o_df, get_full_df)

# from .biomass.growth_summary import get_Biomass_df, get_desired_cycle, get_end_BM

# from .flux.flux_ratio_analysis import get_carbon_allocation_summary
from .drug_combination_response.classification import convert_po_col
from .flux.flux_ratio_analysis import (get_p_o_df, get_full_df)

from .biomass.growth_summary import (get_biomass_df, get_desired_cycle, get_end_BM)
from .flux.flux_snapshot import FluxCompare
# from flux.biomass.pathway_summary import (get_SG_pwy_df, get_pwy_rxn_df)
# from flux.pathway_summary import (get_rct_pathway_df,get_gene_pathway_df, 
#                              get_gene_count_df, get_single_pathway_df)
from .flux.carbon_allocation import (get_carbon_allocation_E_wide, get_fs_change)

from .plot.heatmap import (generate_row_colors, relabel_clustermap, assign_plot_total_E_wide)
from .plot.kde import (get_fs_kde_plot, plot_kde)
from .plot.legend import get_row_color_legend
from .plot.scatter import scatter_xycol_A, scatter_xycol_B


from scarcc.preparation.perturbation.alpha_finder import get_div_obj_df
from scarcc.preparation.metabolic_model import BasicModel
import os
def find_directory(start_directory, directory_name):
    current_directory = os.path.abspath(start_directory)

    while True:
        directory_path = os.path.join(current_directory, directory_name)
        if os.path.isdir(directory_path):
            return directory_path

        # Move up one level
        current_directory = os.path.dirname(current_directory)

        # Check if reached the root directory
        if current_directory == os.path.dirname(current_directory):
            # Directory not found
            return None
data_directory = find_directory(os.path.abspath(__file__), 'models')
E0, S0, all_components = BasicModel(model_directory=data_directory, flux_weighting=True).load_ES_models()

df_list = get_div_obj_df([S0], 0.5, potential_genes=['folA'])
print(df_list)







# Find alpha from 

from dataclasses import dataclass, field
import logging
import pandas as pd
from typing import List, Dict

from scarcc.preparation.perturbation import alter_Sij, iter_species
from scarcc.util import convert_arg_to_list
from .alpha_finder import AlphaFinderConfig

logger = logging.getLogger(__name__)

potential_genes = [
    'glyA', 'gltA', 'tktA', 'dapF', 'dapB', 'acnB', 'pgk', 'talB', 'mrcA', 'pyrE', 'dapD', 'pfkA', 'gdhA', 'folA', 'mrdA', 'thrB',
    'dapA', 'serC', 'argD', 'thrC', 'aceF', 'pykF', 'dadX', 'folC', 'pyrD', 'trpA', 'serB', 'fbp', 'eno', 'pgi', 'pheA',
    'gcvH', 'gnd', 'murA', 'aroA', 'guaB', 'glnA', 'yrbG', 'folP', 'purU', 'serA', 'gltD', 'purT', 'ackA', 'purN', 'rffG',
    'gapA'
]

@dataclass(kw_only=True)
class MonocultureAlphaFinder(AlphaFinderConfig):
    """FBA monoculture alpha finder to restrict biomass production at Normalized target biomass
    *args and **kwargs are passed to the AlphaFinderConfig parent class

    Parameters
    ----------
    model : 'cobra.Model'
        Single cobra model
    exp_leap : float
        Exponential search leap
    is_monoculture : bool
        Whether the alpha search is done in FBA, monoculture
    trace_obj_val : List
        List of objective values result from the search process
    norm_obj_val : float
        Normalized objective value
    response_record : Dict
        Record flux and other summary information under different alpha from the search process
    opt_df : pd.DataFrame
        Detailed data frame to be returned with flux and reaction information for the optimal alpha  

    Attributes
    ----------
    response_record : Dictionary with the following structure
        {current_gene: {model: {'Normalized target biomass': target_obj_val, 
                                'response': {
                                    obj_val_interval: {
                                        'search_alpha' : search_alpha,
                                        'precise_obj_val' : precise_obj_val, 
                                        'interval_abs_diff' : interval_abs_diff,
                                        'target_abs_diff' : target_abs_diff
                                    }
                                }}}}
    
    """
    # TODO: Implementation using linear programming to skip search process

    model : 'cobra.Model' # single cobra model
    exp_leap : float = 2
    is_monoculture : bool = True
    trace_obj_val : List = field(default_factory=list)
    norm_obj_val : float = None # gr_Nomral in coculture case
    response_record: Dict = None # for iterative update in checkerboard
    opt_df : pd.DataFrame = None

    def __post_init__(self):
        if not self.acceptance_threshold_lower:
            self.acceptance_threshold_lower = .85
        if not self.acceptance_threshold_upper:
            self.acceptance_threshold_upper = 1.1
        self.norm_obj_val = self.model.slim_optimize()
        if not self.response_record:
            self.response_record = {self.current_gene: {self.model: {'Normalized target biomass': convert_arg_to_list(self.target_obj_val), 'response': {}}}}

    def fill_response_record(self, obj_val, is_lowest_abs_diff=False):
        """Fill response record with the objective value, if objective value falls into previously visited category, update the record

        Parameters
        ----------
        obj_val : float
            Objective value
        is_lowest_abs_diff : bool
            Whether the objective value has the lowest absolute difference from the target objective value
        """
        obj_val_interval = float(format(obj_val, '.2f'))
        record = self.response_record[self.current_gene][self.model]['response'].get(obj_val_interval) # closest alpha in the obj_interval
        interval_abs_diff = abs(obj_val - obj_val_interval)

        if record and (record['interval_abs_diff'] >= interval_abs_diff): # erase record, update to current alpha
            record = None
        if record and (obj_val_interval > .99) and (record['search_alpha']>self.search_alpha): # IC0 update most large alpha
            record = None

        if self.opt_df is not None:
            previous_min = min([ele['target_abs_diff'] for ele in
                    self.response_record[self.current_gene][self.model]['response'].values()])
            if interval_abs_diff <= previous_min:
                is_lowest_abs_diff = True

        if not record:
            self.response_record[self.current_gene][self.model]['response'][obj_val_interval] = {
                'search_alpha' : self.search_alpha,
                'precise_obj_val' : obj_val, 
                'interval_abs_diff' : interval_abs_diff,
                'target_abs_diff' : abs(obj_val - self.target_obj_val)}
        return is_lowest_abs_diff

    def eval_alpha_fun(self):
        """Evaluate the current alpha and update the optimal dataframe"""
        _, obj_val, summary_df = Sij_biomass(self.model, self.search_alpha, self.current_gene)
        
        if(self.opt_df is None and summary_df['Net_Flux'][0]<1e-8): # reinitialize if the first alpha to search gives zero flux
            self.search_alpha = 1+2e-3
            _, obj_val, summary_df = Sij_biomass(self.model, self.search_alpha, self.current_gene)
        self.trace_obj_val.append(obj_val)
        obj_val = obj_val/ self.norm_obj_val # to normalized 
        self.is_new_ub = (summary_df['Net_Flux'][0]<1e-8 or summary_df['Net_Flux'][0]>100 or 
                        obj_val<self.target_obj_val*self.acceptance_threshold_lower or obj_val>1.2) # overinhibition, search alpha too high 
        
        self.found_alpha = self.is_found(self.search_alpha, self.alpha_lb, self.alpha_ub, self.precision)
        
        # update optimal df
        net_flux_req = (summary_df['Net_Flux'][0]>0 or obj_val>self.norm_obj_val*.1)
        is_lowest_abs_diff = self.fill_response_record(obj_val)

        obj_req = is_lowest_abs_diff
        if (net_flux_req and obj_req) or (self.opt_df is None): # store only qualified alpha
            self.opt_df = summary_df
        
    def out_fun(self):
        """Output the result of the alpha search
        
        Returns
        -------
        Tuple
            Series of alpha, biomass, and the optimal dataframe"""
        opt_df = self.opt_df
        opt_df['is_growth_switch'] = self.classify_growth_switch()
        logger.debug('Gene: %s with alpha %s and obj_val %s', self.current_gene, opt_df['alpha'][0], opt_df[f'biomass'][0])
        opt_df.insert(loc=2, column='Normalized_biomass', value=opt_df['biomass']/self.norm_obj_val)
        opt_df.columns = [f'{self.model.id}_'+element for element in opt_df.columns]
        self.opt_df = opt_df
        return (opt_df[f'{self.model.id}_alpha'], opt_df[f'{self.model.id}_biomass'], opt_df) # alpha_feasible, and upper bound of alpha

def get_summary_df(model, alpha, obj_val, rct_ids = 'DHFR', sol_sol = None) -> pd.DataFrame: 
    """Dataframe summary from the optimization result

    Parameters
    ----------
    model : cobra.Model
    alpha : float
    obj_val : float
        target biomass
    rct_ids : str
        reaction ids
    sol_sol : FBA solution, Optional - flux distribution summary

    Returns
    -------
    pd.DataFrame with the following columns
        - 'alpha' (float): optimal searched alpha
        - 'biomass' (float): biomass yield associated with the perturbation characterized by alpha  
        - 'forward_reactions_id' (str): reaction ids in forward direction
        - 'forward_reactions' (str): reaction names in forward direction
        - 'forward_flux_values' (list): flux values in forward direction
        - 'opposite_reactions' (str): reaction ids in opposite direction
        - 'opposite_flux_values' (list): flux values in opposite direction
        - 'Net_Flux_Boolean' (str): 'Zero Flux' if net flux is zero, 'Net Flux' otherwise
        - 'Net_Flux' (float): net flux
    """
    if type(alpha) != int and type(alpha) != float:
        alpha = str(alpha)

    if sol_sol is None:
        return pd.DataFrame([{'alpha': alpha,'biomass': obj_val}])
    
    rct_dir, rct_rev, flux_dir, flux_rev = list(), list(), list(), list()
    for current_rct_id in rct_ids:
        append_str = str(model.reactions.get_by_id(current_rct_id))
        if ('_v1' not in current_rct_id): # direction align with FVA
            rct_dir.extend([append_str])
            try:
                flux_dir.extend([round(model.reactions.get_by_id(current_rct_id).flux,5)])
            except:
                flux_dir.extend([sol_sol[current_rct_id]])
        else: # direction opposite
            rct_rev.extend([append_str])
            try:
                flux_rev.extend([round(model.reactions.get_by_id(current_rct_id).flux,5)])
            except:
                flux_rev.extend([sol_sol[current_rct_id]])
    net_flux = sum(abs(element) for element in set(flux_dir) | set(flux_rev))
    net_flux_bool = 'Zero Flux' if net_flux ==0 else 'Net Flux'

    summary_df = pd.DataFrame({f'alpha': alpha,
                                f'biomass': obj_val,
                                f'forward_reactions_id': ', '.join(rct_ids),
                                f'forward_reactions': ' '.join(rct_dir),                                    
                                f'forward_flux_values':[flux_dir],
                                f'opposite_reactions': ', '.join(rct_rev),     
                                f'opposite_flux_values':[flux_rev],
                                f'Net_Flux_Boolean':[net_flux_bool],
                                f'Net_Flux':[net_flux]}
                                )
    return(summary_df)

def Sij_biomass(model: "cobra.Model", alphas: float = 1, genes: str = 'folA', slim_opt: bool = False):
    """Get biomass form FBA optimization with given gene and alpha
    
    Parameters
    ----------
    model : cobra.Model
    alphas : float
    genes : str
    slim_opt : bool
        If True, use slim_optimize
        If False, use optimize
        
    Returns
    -------
    float if slim_opt is True
        Objective value
    list(alpha, biomass, summary_df) if slim_opt is False
    """
    with model:
        rct_ids = alter_Sij(model, alphas, genes) # same

        if (slim_opt == False):
            sol_sol = model.optimize()
            obj_val = sol_sol.objective_value 
            summary_df = get_summary_df(model, alphas, obj_val, rct_ids, sol_sol.fluxes)  
            return(alphas, obj_val, summary_df) 
        else:
#             return(pd.DataFrame([{f'biomass': model.slim_optimize()}]))
            return(model.slim_optimize())

def get_single_div_obj_df(model: "cobra.Model", target_obj_val: float, potential_genes: list, precision: int = 3, search_alpha: float= 1.02,
                        alpha_table: pd.DataFrame = None, acceptance_threshold_upper: float = 1.1, acceptance_threshold_lower: float=.95) -> pd.DataFrame: 
    """Get the objective value for each gene in a single model
    If alpha_table is not provided, perform a search for the optimal alpha 
    and record the corresponding objective value
    
    Parameters
    ----------
    model : cobra.Model
    target_obj_val : float
        The desired target objective value
    potential_genes : list
        List of potential genes to generate alpha_table
    precision : int
        The precision of the alpha
    search_alpha : float
        initial alpha to start search
    alpha_table : pd.DataFrame, Optional
        The alpha table
    acceptance_threshold_upper : float
        The upper acceptance threshold for the target objective value
    acceptance_threshold_lower : float
        The lower acceptance threshold for the target objective value

    Returns
    -------
    pd.DataFrame
        summary_df        
    """
    obj_div_df = pd.DataFrame() # obj_val: biomass/ growth rate

    for current_gene in potential_genes: # iter gene 
        logger.debug('%s in cal', current_gene)
        print(current_gene, 'in cal')
        with model:
            if alpha_table is None: 
                if not search_alpha: search_alpha = 1.02
                # AlphaFinder class for every model and SG
                AF = MonocultureAlphaFinder(model=model,
                            search_alpha = search_alpha,
                            current_gene = current_gene, 
                            target_obj_val = target_obj_val, 
                            exp_leap=3,
                            precision=precision,
                            acceptance_threshold_upper = acceptance_threshold_upper,
                            acceptance_threshold_lower = acceptance_threshold_lower)
                alpha, obj_value, temp_df = AF.find_feasible_alpha()
            else:
                alpha = alpha_table.loc[current_gene,f'{model.id}_alpha'] # get alpha for corresponding gene from alpha_df
                alpha, obj_value, temp_df = Sij_biomass(model, alpha,
                                                    genes = current_gene)
            temp_df['Gene_inhibition'] = current_gene 
            obj_div_df = pd.concat([obj_div_df, temp_df.set_index('Gene_inhibition')],axis=0)
    return obj_div_df

def get_div_obj_df(model_list: List["cobra.Model"]=None, target_obj_val: float=None, potential_genes: List=potential_genes,
                    precision: float=3, detailed_alpha_table: bool=False, div_obj_df: pd.DataFrame=None):
    """Get the objective value for each gene in each model in model_list
    
    Parameters
    ----------
    model_list : List
        List of cobra models
    target_obj_val : float
        The desired target objective value
    potential_genes : List
        List of potential genes 
    precision : float
        The precision of the alpha
    detailed_alpha_table : bool
        boolean to return detailed alpha table or not
    div_obj_df : pd.DataFrame, Optional
    
    Returns
    -------
    pd.DataFrame
        summary_df if detailed_alpha_table is False
        alpha and normalized_biomass columns of summary_df otherwise
    """
    if div_obj_df is None:
        alpha_obj_df_list = iter_species(model_list, get_single_div_obj_df,
                                target_obj_val=target_obj_val, potential_genes=potential_genes, precision=precision)
        result_df = pd.concat(alpha_obj_df_list, axis=1)
    else:
        result_df = div_obj_df
    if detailed_alpha_table:
        return result_df
    return result_df.filter(regex='alpha|Normalized_biomass')
    


import logging
import pandas as pd

from scarcc.preparation.perturbation import alter_Sij, get_alphas_from_tab
# from scarcc.preparation.perturbation import iter_species
from scarcc.preparation.metabolic_model import BasicModel

logging.basicConfig(filename='test_stoi_scaling.log', 
                    level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                    filemode='w')

E0, S0, all_components = BasicModel(flux_weighting=True).load_ES_models()
alpha_table = pd.read_csv('../Data/alpha_table_m1.csv', index_col=0)

def iter_species(models,f,*args,**kwargs): 
    r_object = list()
    if type(models) is not zip:
        for model in models:
            print(*args,**kwargs)

            r_object.append(f(model,*args,**kwargs))
    else:
        for model, *extra_objects in models: 
            r_object.append(f(model,extra_objects,*args,**kwargs))
    return(r_object) 

def change_model_stoichiometry(model, alphas):
    with model:
        changed_rxns = alter_Sij(model, alphas=alphas, genes='folA', ko=False)
        print(changed_rxns)
        logging.debug('For model %s with alpha %s, Scaled reactions: %s', model.id, alphas[0],'\n'.join([str(model.reactions.get_by_id(rxn)) for rxn in changed_rxns]))
    
species_list = [E0, S0]
alpha_list = [get_alphas_from_tab(model, ['folA'], alpha_table) for model in species_list]
print(alpha_list)
zipped_sp_alpha = zip([E0, S0], alpha_list)
iter_species(zipped_sp_alpha, change_model_stoichiometry)


import os
import json
import numpy as np

from dataclasses import dataclass, field
from typing import List
import pandas as pd
import concurrent.futures

from scarcc.util import (convert_arg_to_list, rename_columns)
from scarcc.data_analysis.biomass.growth_summary import get_desired_cycle
# from scarcc.sim_engine.simulation_workflow import extract_biomass_flux_df
from .alpha_finder import AlphaFinderConfig

@dataclass(kw_only=True)
class CocultureAlphaFinder(AlphaFinderConfig): # scale normal & knockout to 1-0
    model: List['cobra.Model']
    search_alpha : float
    current_gene : str
    target_obj_val : float = 0.5 
    alpha_table : pd.DataFrame = None
    precision : int = 1 # precision of alpha
#     alpha_table : pd.DataFrame = alpha_table # init AFConfig
    trace_biomass : List = field(default_factory=list)
    trace_obj_val : List = field(default_factory=list) # TODO: combine all trace_XX & convert to dict, follow format of response_record in MonocultureAlphaFinder
    trace_alpha : List = field(default_factory=list)
    gr_ko : float = None
    n_dir : str = ''
    opt_alpha_table: pd.DataFrame = field(default_factory=pd.DataFrame)
    nxt_alpha_table: pd.DataFrame = field(default_factory=pd.DataFrame)
    ko_intercepted : None = None
    acceptance_threshold_lower : float = .95
    acceptance_threshold_upper : float = 1.05
    gr_Normal : float = None
    carbon_source_val : float = 5e-3
    add_nutrient_val : float = 10 # DO NOT set Met limiting in batch culture 
    is_growth_switch = False
    initial_pop : float = 1e-8
    p : None = None
    obj_style: str = 'MAX_OBJECTIVE_MIN_TOTAL'
    data_path = None # replace ./Data/

    # add_nutrient_val : float = field(default_factory=[.08])
    
    def __post_init__(self):
        self.is_monoculture = False
        self.eval_alpha_fun = self.eval_alpha_fun 
        self.get_alpha_use()
        self.E0 = [ele for ele in convert_arg_to_list(self.model) if ele.id == 'E0'][:1]
        # self.gr_Normal = 0.00064644123538125 # ?check effect of lcts conc changed in Div_COMETS
        self.alpha_pair : pd.DataFrame = None
        if not self.exp_leap:
            self.exp_leap = 1.3
        self.E0 = self.model[0]
        self.S0 = self.model[1]
        self.out_fun = self.out_fun
        self.calculate_gr_Normal()
    
    def calculate_gr_Normal(self):
        if not self.gr_Normal:
            if self.current_gene == 'Normal':
                self.gr_Normal = self.calculate_gr_ko()
            else: self.gr_Normal = 0.28672261
        return float(self.gr_Normal)

    def calculate_gr_ko(self):
        print('calculating ko gr')
        self.ko_intercepted = False
        
        if self.current_gene in ['folP', 'folA', 'dadX', 'pgk', 'acnB']:
            self.gr_ko = 0
            return self.gr_ko
        
        # alpha_in_table = self.alpha_table.loc[self.current_gene, self.E0.id] # E0 col
        full_df, *_ =  extract_biomass_flux_df(self.current_gene,n_dir=self.n_dir,alpha_table=self.alpha_table,
                                E0=self.E0, S0=self.S0, mono=False, p=self.p, return_sim=True, ko=True, 
                                carbon_source_val=self.carbon_source_val, add_nutrient_val=self.add_nutrient_val,
                                initial_pop=self.initial_pop, obj_style=self.obj_style)
        
        coculture_biomass_df = full_df.iloc[:,[0]]
        if self.current_gene == 'Normal':
            coculture_biomass_df.to_csv(os.path.join(self.data_path, f'Normal_biomass_{str(self.carbon_source_val)}.csv'))
        
        coculture_biomass_df.columns = rename_columns(coculture_biomass_df)
        
        # desired_cycle = get_desired_cycle(coculture_biomass_df, scale_diff=0.05)
        # # return desired_cycle, full_df, coculture_biomass_df
        # growth_phase = desired_cycle.growth_phase[0]
        # gr_ko = (coculture_biomass_df.loc[growth_phase[1]] - coculture_biomass_df.loc[growth_phase[0]])/(growth_phase[1]-growth_phase[0])
        # gr_ko = float(gr_ko)
        
        gr_ko = self.cal_fitted_gr(coculture_biomass_df)
            
        self.cocdf = coculture_biomass_df 
        # not setting alpha, not upper bound
        self.gr_ko = float(gr_ko)
        
        if self.current_gene == 'Normal':
            print('set_normal')
            self.gr_Normal = gr_ko
        elif self.gr_ko > self.gr_Normal*.9:
            print('intercept')
            self.ko_intercepted = True
            self.opt_alpha_table = self.alpha_table.loc[[self.current_gene]]
        else: print('NNNOT intercept')
        
        print(self.opt_alpha_table)
        
        print(f'self.gr_ko > self.gr_Normal: {self.gr_ko},{self.gr_Normal*.85}')
        print(' gr_ko: ', gr_ko)
        self.trace_biomass.append(full_df.add_suffix(f'_ko'))
        return gr_ko
    
    def get_alpha_use(self):
        if self.current_gene == 'Normal':
            return 0
        Ealpha = self.alpha_table.loc[self.current_gene].iloc[0]
        Salpha = self.alpha_table.loc[self.current_gene].iloc[1]
        self.alpha_use = Ealpha if Ealpha<1e5 else Salpha
        if self.alpha_use > 1e4:
            self.alpha_use = 1.04
            self.exp_leap = 3
        
        if not self.search_alpha and (self.current_gene != 'Normal'):
                self.search_alpha = (self.alpha_use-1)/2+1 if self.alpha_use > 2 else self.alpha_use
        return None
    
    def generate_next_fixed_ratio_table(self):     
        multiplier = (self.search_alpha-1)/(self.alpha_use-1)
        nxt_alpha_table =  (self.alpha_table.loc[[self.current_gene]]-1)*multiplier + 1 # ratio offset by 1, minimum is alpha=1
        # trace alpha table change and corresponding gr
        # self.trace_alpha_table.append(nxt_alpha_table.to_dict())
        self.nxt_alpha_table = nxt_alpha_table
        return nxt_alpha_table

    def cal_fitted_gr(self, coculture_biomass_df):
        def logistic(x, saturation, growth_rate, inflection_point, initial_pop):
            return (saturation / (1 + np.exp(-growth_rate * (x - inflection_point)))) + initial_pop
        
        x = coculture_biomass_df.index
        y = coculture_biomass_df.iloc[:, 0] # E0 column
        # y = coculture_biomass_df # E0 column
        if (float(y.tail(1)) < float(y.head(1))+5e-8):
            growth_rate = 0
            return growth_rate
            
        desired_cycle = get_desired_cycle(coculture_biomass_df, scale_diff=0.05)
        
        growth_phase = desired_cycle.loc[self.current_gene, 'growth_phase']
        bool_growing = desired_cycle.loc[self.current_gene, 'bool_growing']
        c_max_gr = desired_cycle.loc[self.current_gene, 'c_max_gr']
        saturation = float(coculture_biomass_df.loc[growth_phase[1]])
        init_gr = .23
        p0 = [saturation, init_gr, c_max_gr, self.initial_pop]

        print('growing: ',bool_growing)
        if (bool_growing == False) or (bool_growing <1):
            popt, pcov = curve_fit(logistic, x, y, p0=p0)
            print('logi')
            
            if self.current_gene =='Normal':
                print(popt)
            growth_rate = popt[1]
        else:
            print('lolin')
            popt, pcov = np.polyfit(x, np.log(y), 1) # [B,A] -- log(y) = A+Bx
            growth_rate = popt
        if self.current_gene == 'Normal':
            print('popt', popt)
        return growth_rate
        
    def cal_std_gr(self, coculture_biomass_df, gr_Normal):
#         desired_biomass_df = pd.DataFrame(get_cycle_max_gr(coculture_biomass_df), index=['max_gr', 'start', 'end', 'bool_growing'])
        gr = self.cal_fitted_gr(coculture_biomass_df)
        standardized_gr = (float(gr) - self.gr_ko)/(gr_Normal-self.gr_ko) # offset scale self.gr_ko then scale
        if standardized_gr > 1:
            self.exp_leap = 3 # for coculture gr inherently > 1, require large leap
        
        # print('standardized gr: ', standardized_gr)
        self.trace_obj_val.append(standardized_gr)
        print('---GR, stdGR__:',gr, standardized_gr)
        return standardized_gr
    
    def eval_alpha_fun(self, alpha_overwrite=None): # first, search alppha = alpha_table alpha
        if self.ko_intercepted is None:
            _ = self.calculate_gr_ko() # also catch gr_ko ~ gr_Normal- Gene with nonessential reactions
        
        if alpha_overwrite:
            self.search_alpha = alpha_overwrite
        nxt_alpha_table = self.generate_next_fixed_ratio_table()
        # self.trace_alpha_table.append(nxt_alpha_table.to_dict())

        full_df, out_dict, co_sim = extract_biomass_flux_df(self.current_gene,n_dir=self.n_dir,alpha_table=self.nxt_alpha_table,
                                E0=self.E0, S0=self.S0, mono=False, p=p, return_sim=True, ko=False, 
                                carbon_source_val=self.carbon_source_val, add_nutrient_val=self.add_nutrient_val,
                                initial_pop=self.initial_pop, obj_style=self.obj_style)
        # get_BM_df(self.current_gene,n_dir=self.n_dir,alpha_table=nxt_alpha_table,mono=False, return_sim=True)
        self.co_sim = co_sim
        
        self.trace_biomass.append(full_df.add_suffix(f'_{self.i_iter}'))
        
        coculture_biomass_df = full_df.iloc[:,[0]]
        target_gr = self.target_obj_val
        std_gr = self.cal_std_gr(coculture_biomass_df, self.gr_Normal) # ?check this
#         std_gr = self.cal_std_gr(coculture_biomass_df, 1.3)

        self.is_new_ub = (std_gr < target_gr)# sharp bound, although lower estimate, no search for higher alpha if just meet
        self.converge_alpha = self.is_found(self.search_alpha, self.alpha_lb, self.alpha_ub, self.precision)
        
        # self.found_alpha = self.found_alpha or 
        #     (self.std_gr > self.target_obj_val*acceptance_threshold_upper)

        self.obj_found = ((std_gr > self.target_obj_val*self.acceptance_threshold_lower) and 
            (std_gr < self.target_obj_val*self.acceptance_threshold_upper)) 
        
        bool_gr_coculture_exceed_limit = (std_gr > 1)
        self.found_alpha = self.converge_alpha or self.obj_found or self.classify_growth_switch()

        if self.is_new_ub:
            self.nxt_alpha_table = nxt_alpha_table
        print('searc, found, new_un, lu')
        print(self.search_alpha, self.found_alpha, self.is_new_ub, self.alpha_lb, self.alpha_ub)
        print('convg, obj_f, exceed, gr_sw')
        print(self.converge_alpha, self.obj_found, bool_gr_coculture_exceed_limit, self.classify_growth_switch())
        # update optimal df
        
        obj_req = ((std_gr > self.target_obj_val*0.95) and (std_gr < self.target_obj_val*1.1)  # harsh lb
                    or self.is_growth_switch)
        
        temp_df = pd.DataFrame.from_dict(
            {self.current_gene :
                {'std_growth_rate' : std_gr,
                'lb_ub' : (self.alpha_lb, self.alpha_ub),
                'converge_alpha' : self.converge_alpha,
                'obj_found': self.obj_found,
                'bool_gr_coculture_exceed_limit': bool_gr_coculture_exceed_limit,
                'classify_growth_switch': self.is_growth_switch,
                'Non_essential': self.ko_intercepted
                }}, 
                orient='index')
        nxt_alpha_table = pd.concat([nxt_alpha_table, temp_df], axis=1)
        nxt_alpha_table.index.name='Gene_inhibition'
        self.trace_alpha.append(nxt_alpha_table.to_dict())
        
        
        if obj_req or (self.opt_df is None) or self.converge_alpha: # store only qualified alpha
            if self.opt_df is not None:
                nxt_alpha_table['obj_req_satisfied'] = True
            self.opt_df = [full_df, nxt_alpha_table]
            # self.opt_alpha_table = pd.concat([nxt_alpha_table, temp_df], axis=1)
            self.opt_alpha_table = nxt_alpha_table
        return full_df 
    
    def out_fun(self):
        result = {}
        for i in range(len(self.trace_obj_val)):
            result[f'iter_{i+1}'] = {'obj_val': self.trace_obj_val[i], 'alpha_table': self.trace_alpha[i]}
        
        biomass_df = pd.concat(self.trace_biomass, axis=1) if self.trace_biomass else pd.DataFrame()
        return biomass_df, self.opt_alpha_table, {self.current_gene : result}
#         return 'Temp Done' if (n_iter<=10 or n_iter ==2) else 'end at: '+ str(n_iter)

##--------Perform Coculture Secrch---------##

def coculture_search_job(**kwargs):    
    AF = CocultureAlphaFinder(**kwargs)
    AF.calculate_gr_ko()
    return AF.find_feasible_alpha()

def run_coculture_search_mp(potential_genes, data_path, filename, n_processor, **kwargs):
    def save_trace_biomass(trace_biomass):
        trace_biomass = pd.concat(trace_biomass,axis=1) # alpha_table
        trace_biomass.columns = rename_columns(trace_biomass)
        trace_biomass.to_csv(os.path.join(data_path, f'biomass_{filename}.csv'))
    
    result_list = list()
    with concurrent.futures.ProcessPoolExecutor(n_processor) as executor:
        for i, current_gene in enumerate(potential_genes):
            future = executor.submit(coculture_search_job, current_gene=current_gene, n_dir=i, **kwargs)
            result_list.append(future)
    
    trace_biomass, opt_alpha_list, result_dict_list = zip(*[r.result() for r in result_list])
    opt_alpha = pd.concat(opt_alpha_list) # alpha_table
        
    opt_alpha.columns = rename_columns(opt_alpha)
    opt_alpha.to_csv(os.path.join(data_path, f'alpha_table_{filename}.csv'))
    print(opt_alpha, result_dict_list)
    
    trace_dict = {k: v for d in result_dict_list for k, v in d.items()}
    with open(os.path.join(data_path, f"trace_record_{filename}.json", "w")) as outfile:
        json.dump(trace_dict, outfile)

    save_trace_biomass(trace_biomass)
    print('finished alpha search')
    return opt_alpha_list, trace_dict


from .perturbation.alpha_finder.alpha_finder import AlphaFinderConfig
from .perturbation.alpha_finder.monoculture import MonocultureAlphaFinder, get_div_obj_df
from .perturbation.alpha_finder.coculture import CocultureAlphaFinder, run_coculture_search_mp



import pandas as pd
import re
from scarcc.preparation.metabolic_model import get_component
from scarcc.data_analysis import convert_po_col
# from scarcc.data_analysis.

def get_total_carbon(rxn, model, all_components, is_substrate=True,  carbon_detail=False, atp_only=False, atp_exclude=True): # TODO: exclude EX_met__L_e
    def get_n_carbon(formula):
        if formula in ['CO2', 'CH1O2']:
            return 1
        value = re.findall(r'C(\d+)', formula)
        numeric_value = int(value[0]) if value else 0
        return(numeric_value)
    
    def eval_include_in_dict(metabolite, factor, is_substrate=True):
        if atp_exclude and metabolite.id == 'atp_c':
            return False
        if rxn.id == 'BIOMASS_Ec_iML1515_core_75p37M' and atp_only and metabolite.id != 'atp_c':
            return False
        return (factor<0) == is_substrate and get_n_carbon(metabolite.formula)!=0
    
    if isinstance(rxn, str):
        rxn = get_component(model, rxn, all_components)
    d = {m.id: {'formula': m.formula,
                'formula_carbon': get_n_carbon(m.formula),
                'factor': factor,
                'carbon': factor * get_n_carbon(m.formula),}
                for m, factor in rxn.metabolites.items() if eval_include_in_dict(m, factor, is_substrate)}
    carbon_only = {k:v['carbon'] for k,v in d.items()}
    if carbon_detail:
        return carbon_only
    carbon_total = sum(carbon_only.values())
    return carbon_total

def get_carbon_allocation_summary(s, model, all_components, detailed=False, carbon_dict= {}, format=True): # accept series as input and return allocation df   
    for rxn, flux_value in s.filter(regex='EX_.*e$|BIOMASS_Ec_iML1515_core_75p37M$').dropna().items():
        
        carbon_per_flux = get_total_carbon(rxn, is_substrate=True, model=model, all_components=all_components, atp_exclude=True)

        carbon_dict.update({rxn: {
                            'carbon_per_flux': carbon_per_flux,
                            'flux_quantity': flux_value,
                            'carbon_exchange': float(flux_value * carbon_per_flux),
                            }})
    if detailed:
        return pd.DataFrame(carbon_dict).T

    carbon_allocation = pd.DataFrame({rxn: {'total_carbon': v['carbon_exchange']}
                        for rxn, v in carbon_dict.items() if abs(v['carbon_exchange'])>1e-4}).T
    normalize_carbon = carbon_allocation.loc['EX_lcts_e','total_carbon']
    carbon_allocation['percent'] = (carbon_allocation.total_carbon/normalize_carbon)
    waste_portion = -1-carbon_allocation.query('percent<0').percent.sum()
    if abs(waste_portion)<0.01: # if waste portion is too small, ignore it
        waste_portion = 0

    waste_row = pd.DataFrame([-1*waste_portion*normalize_carbon, waste_portion] # inclusion of waste product
                            ,columns=['Waste'], index=carbon_allocation.columns).T
    
    carbon_allocation = pd.concat([carbon_allocation, waste_row])
    if format:
        carbon_allocation['percent'] = carbon_allocation['percent'].apply(lambda x: f'{x:.2%}')
    carbon_allocation['Gene_inhibition'] = s.name if isinstance(s.name, str) else s.name[0]
    carbon_allocation.index.name = 'reaction'
    return carbon_allocation

def get_carbon_allocation_E_wide(E0, all_components, additive_threshold=0.05, flux_analysis_file='./Data/flux_analysis_m1.csv', gr_file='./Data/gr_DG_m1_normalized.csv', end_BM=None):
    # flux_analysis_full = pd.read_csv('./Data/flux_analysis_m1.csv', index_col=0)
    flux_analysis_full = pd.read_csv(flux_analysis_file, index_col=0)
    co_E = flux_analysis_full.query('Species=="E0"')
    if 'culture' in co_E.columns:
        co_E = co_E.query('culture=="coculture"')
    # if 'XG'in co_E.columns:
    #     co_E = co_E.query('XG=="DG"')
    gr_df = pd.read_csv(gr_file, index_col=0)
    # gr_df = pd.read_csv('./Data/gr_DG_m1_normalized.csv', index_col=0)
    carbon_allocation_E = pd.concat(co_E.apply(lambda x: 
                                            get_carbon_allocation_summary(x, E0, all_components,format=False), axis=1)
                                                .tolist())
    carbon_allocation_rxns = get_carbon_allocation_summary(co_E.iloc[0], E0, all_components).index # carbon allocation fluxes

    carbon_allocation_E_wide = carbon_allocation_E.reset_index().pivot(index='Gene_inhibition',columns='reaction',values=['total_carbon', 'percent'])
    # carbon_allocation_E_pivot = carbon_allocation_E.pivot(index='Gene_inhibition',columns='reaction',values=['total_carbon', 'percent'])
    carbon_allocation_E_wide.columns = ['_'.join([col[0], col[1]]) for col in carbon_allocation_E_wide.columns]
    carbon_allocation_E_wide = carbon_allocation_E_wide.merge(co_E[['XG']], left_index=True, right_index=True) # only E0
    partial_convert_po_col = lambda x: convert_po_col(x, additive_threshold=additive_threshold)
    carbon_allocation_E_wide['Drug_comb_effect_coc'] = partial_convert_po_col(gr_df['po_diff_E0_coculture'])
    carbon_allocation_E_wide['Drug_comb_effect_Emono'] = partial_convert_po_col(gr_df['po_diff_E0_monoculture'])
    carbon_allocation_E_wide['Drug_comb_effect_Smono'] = partial_convert_po_col(gr_df['po_diff_S0_monoculture'])

    # return carbon_allocation_E_wide
    if 'BM_consortia_frac_binned' not in carbon_allocation_E_wide.columns and end_BM is not None:
        carbon_allocation_E_wide = carbon_allocation_E_wide.merge(end_BM.query('Species=="E0"').BM_consortia_frac_binned, left_index=True, right_index=True)
    carbon_allocation_E_wide = carbon_allocation_E_wide.reset_index('Species', drop=True)
    # carbon_allocation_E_wide.to_csv('./Data/carbon_allocation.csv')
    return carbon_allocation_E_wide

## Difference in normalized carbon flux for double gene compared to First and second 
def get_percent_cols(df, col_suffix=None, col_prefix='percent'):
    df = df.filter(regex=col_prefix)
    if not col_suffix:
        return df
    df = df.filter(like=col_suffix) 
    df.columns = df.columns.str.replace(col_suffix,'')
    return df

def get_fs_change(gr_path, carbon_allocation_E_wide, additive_threshold=0.05):
    def get_sub_fs_change(fs_df, col_suffix='_First', col_prefix='percent'):
        fs = get_percent_cols(fs_df, col_suffix, col_prefix=col_prefix)
        # return carbon_allocation_E_wide
        fs_change = fs - get_percent_cols(carbon_allocation_E_wide.loc[fs.index], col_suffix=None, col_prefix=col_prefix) # change compared to single gene
        # fs_change = fs_change.merge(carbon_allocation_E_wide[['Drug_comb_effect_coc']], left_index=True, right_index=True).dropna(axis=0, how='all')
        return fs_change
    
    gr_df = pd.read_csv(gr_path, index_col=0)
    fs_df = carbon_allocation_E_wide.merge(gr_df[['First_gene', 'Second_gene']], left_index=True, right_index=True)

    fs_df = (fs_df.merge(carbon_allocation_E_wide, left_on='First_gene', right_index=True, suffixes=['','_First']) # add corresponding Frist and Second flux data
            .merge(carbon_allocation_E_wide, left_on='Second_gene', right_index=True, suffixes=['','_Second']))

    fs_change = pd.concat([get_sub_fs_change(fs_df, '_First'), get_sub_fs_change(fs_df, '_Second')])
    fs_change = fs_change.sort_index()
    fs_change['Nth_gene'] = list(['First','Second']*len(fs_change.index.unique()))
    # if 'Drug_comb_effect_coc' in fs_change.columns:
    #     fs_change = fs_change.drop(['Drug_comb_effect_coc'], axis=1)
    fs_change = fs_change.merge(carbon_allocation_E_wide.loc[:,'Drug_comb_effect_coc':], left_index=True, right_index=True)
    
    return fs_change


from setup import convert_arg_to_list
import itertools
from flux_snapshot import get_SG_DG

def plot_coculture(Biomass_df, gene_list,culture ='coculture', species_id_list = ['E0', 'S0']):
    species_id_list = convert_arg_to_list(species_id_list)

    if not isinstance(species_id_list[0], str):
        species_id_list = [m.id for m in species_id_list]
    colnames = list()
#     for model, current_gene in itertools.product([E0, S0],convert_arg_to_list(gene_list)):
    for model_id, current_gene in itertools.product(species_id_list,convert_arg_to_list(gene_list)):
        colnames.append('_'.join([model_id, current_gene, culture]))
    
    Biomass_df[colnames].plot()
    return None

# def plot_SG_DG_compare(Biomass_df, SG, DG_list, **kwargs):
#     SG_DG = get_SG_DG(DG_list, explode=True)
#     SG=convert_arg_to_list(SG)
#     DG = list(SG_DG.query('SG in @SG').index)
#     if DG:
#         print(SG, 'gcomb growth curve')
#         plot_coculture(Biomass_df, list(itertools.chain(*[DG,SG, ['Normal']])), **kwargs)
#     else: print(SG, 'do not have gcomb')

def plot_SG_DG_compare(Biomass_df, SG, DG_list, **kwargs):
    DG_list = convert_arg_to_list(DG_list)
    SG_DG = get_SG_DG(DG_list, explode=True)
    if SG is None:
        SG = list(itertools.chain(*[gcomb.split('.') for gcomb in DG_list]))
    else:
        SG=convert_arg_to_list(SG)

    DG = list(SG_DG.query('SG in @SG').index)
    if DG:
        print(list(itertools.chain(*[DG,SG, ['Normal']])))
        plot_coculture(Biomass_df, list(itertools.chain(*[DG,SG, ['Normal']])), **kwargs)
    else: print(SG, 'do not have gcomb')
# plot_coculture(Biomass_df, ['murA', 'murA.pyrD'],culture='monoculture')
# plot_SG_DG_compare(SG, culture='monoculture',species_list=['E0'])


import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

def get_row_color_legend():
    color_label = dict()
    color_label['Drug Combination Effect'] = {'Antagonistic': '#90fda9', 'Additive': '#ffff84', 'Synergistic': '#fed0fc', '': 'white'} # place holder
    color_label['rel_abund'] = {'E Dominance (0.7 ≤ Rel. Abund. < 0.75)': '#4622fc', 'E Dominance (0.5 ≤ Rel. Abund. < 0.7)': '#9d22fc', 
                                'S Dominance (Rel. Abund. > 0.5 for S)': '#da52d5', 'No Growth': 'grey'}
    color_label['all_label'] = {}
    color_label['all_label'].update(color_label['Drug Combination Effect'])
    color_label['all_label'].update(color_label['rel_abund'])

    patches = [mpatches.Patch(color=color, label=label) for  label, color in color_label['all_label'].items()]

    plt.figure(figsize=(8,1.46))
    plt.legend(handles=patches, loc='upper left', title='Drug Combination Effect       Species Relative Abundance (Rel. Abund.)      ', ncol=2,frameon=False)
    plt.gca().set_axis_off()
    plt.show()



from setuptools import setup, find_packages

setup(
    name='scarcc',
    version='0.0.10',
    url='https://github.com/TFwongw/scarccpy',
    author='Thomas Wong',
    author_email='wong0755@umn.edu',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(where="src"),
    package_dir = {"": "src"},
    install_requires=[
        'cobra',
        'pandas<2',
    ],
)


"""Summary statistics from biomass growth record"""

import pandas as pd
from typing import List, Union
from scarcc.util import convert_arg_to_list
from ..flux.flux_snapshot import set_GI_SP_as_MI

def turn_to_normal_colname(colname):
    colname = colname.split('_')
    new_colname = '_'.join([colname[0], 'Normal', colname[2]])
    return new_colname

def convert_col_to_gene_inhibition(s):
    if len(s.split('.')) <=2:
        _, gcomb, _ = s.split('_')
        return gcomb
    _, gcomb, _, lv_pairs = s.split('_')
    return gcomb + '_' + lv_pairs

def get_biomass_df(file_path_list: Union[str, List]):
    """read biomass_df from file"""
    biomass_df =  pd.concat(
            [pd.read_csv(file, index_col='cycle')
            for file in convert_arg_to_list(file_path_list)]
        ,axis=1)
    if biomass_df.columns.str.contains('Normal').any():
        return biomass_df
    
    normal_columns = [ele for ele in biomass_df.columns if '0.0' in ele]
    normal_columns_rename_dict = {col: turn_to_normal_colname(col) for col in normal_columns}
    biomass_df.rename(columns=normal_columns_rename_dict, inplace=True)
    return biomass_df

def get_XG_cycle_from(desired_cycle):
    """separate SG and DG cycle from desired_cycle"""
    if len(desired_cycle.index[-1].split('.')) <=2:
        SG_cycle = desired_cycle.loc[[len(ele.split('.')) ==1 for ele in desired_cycle.index]]
        DG_cycle = desired_cycle.loc[[len(ele.split('.')) ==2 for ele in desired_cycle.index]]
    else:
        SG_cycle = desired_cycle.loc[['0' in ele for ele in desired_cycle.index]]
        DG_cycle = desired_cycle.loc[['0' not in ele for ele in desired_cycle.index]]
    SG_cycle.index.name='SG'
    DG_cycle.index.name='DG'
    SG_cycle.columns.name=None
    DG_cycle.columns.name=None
    return SG_cycle, DG_cycle

def search_gr_cycle_with_biomass(df_search, biomass_values):
    """search for cycle with biomass value
    biomass value at half maximum represents middle of exponential phase
    
    Parameters
    ----------
    df_search: pd.Series
        biomass record
    biomass_values: list
        biomass values to search

    Returns
    -------
    list of boolean if biomass record is larger than queried biomass value
    """
    return [df_search[df_search >= biomass_value].idxmin()
                for biomass_value in list(biomass_values)]

def get_maximum_growth_cycle(desired_biomass, scale_biomass_diff=0.1):
    """get summary statistics form biomass record
    
    Parameters
    ----------
    desired_biomass: pd.DataFrame
        biomass record
    scale_biomass_diff: float
        scale for biomass difference(max-min) to account for noise, 
        delayed start and advanced end cycle for exponential phase
        
    Returns
    -------
    c_max_gr: cycle at middle of exponential phase
    start: growth start cycle
    end: growth end cycle
    bool_growing: boolean if growing, cell not reach stationary phase
    """
    c_max_gr = desired_biomass.iloc[1]+ (desired_biomass.iloc[-1] - desired_biomass.iloc[1])/2
    bool_growing = ((desired_biomass.iloc[-1]-desired_biomass.iloc[-5])/desired_biomass.iloc[-1]).apply(lambda x: x > 1e-10)
    for k, bool_grow in bool_growing.items():
        if bool_grow:
            c_max_gr[k] = desired_biomass[k].iloc[-6]
    biomass_diff = (desired_biomass.iloc[-1]-desired_biomass.iloc[0])
    start = desired_biomass.iloc[0] + biomass_diff*scale_biomass_diff
    end = desired_biomass.iloc[0] + biomass_diff*(1-scale_biomass_diff)
    return c_max_gr, start, end, bool_growing

def get_desired_cycle(biomass_df, log_step=5, scale_biomass_diff=0.1):
    """get desired growth cycle from biomass record
    
    Parameters
    ----------
    biomass_df: pd.DataFrame
        biomass record
    log_step: int
        log step of flux record
    scale_biomass_diff: float
        argument for get_maximum_growth_cycle, refined exponential phase
        
    Returns
    -------
    desired_cycle: pd.DataFrame
        summary statistics from biomass record
    """
    def correct_cycle(cycle): # round to nearest available flux log time point
        if cycle < log_step:
            return log_step
        return round(cycle / log_step) * log_step

    def get_growth_phase_length():
        return ((desired_cycle['end'] - desired_cycle['start'])*(1-desired_cycle.bool_growing) + # if not growing, not changing growth phase length
                1e4*(desired_cycle.bool_growing)) #if growing, set growth length to 999
    
    def split_index_to_cols(df):
        items = df.index.str.split('_')
        columns = ['Species', 'Gene_inhibition', 'culture']
        if len(items[0]) > 3:
            columns.extend(['alpha_lv_pairs'])
        return pd.DataFrame(items.tolist(), index=df.index, columns=columns)
    desired_biomass_df = pd.DataFrame(get_maximum_growth_cycle(biomass_df, scale_biomass_diff=scale_biomass_diff)
                                    , index=['c_max_gr', 'start', 'end', 'bool_growing'])
    
    desired_cycle = (desired_biomass_df.iloc[:-1]
                .apply(lambda x: 
                        search_gr_cycle_with_biomass(biomass_df.loc[:,x.name],x))
                .T)
    desired_cycle['bool_growing'] = desired_biomass_df.T.bool_growing
    desired_cycle['cycle_max_gr'] = desired_cycle['c_max_gr'].apply(correct_cycle) # -> cycle_max_gr
    desired_cycle['growth_phase'] = desired_cycle[['start', 'end']].values.tolist()
    desired_cycle['growth_phase_length'] = get_growth_phase_length()
    desired_cycle['end_cycle'] = biomass_df.index[-1]//5*5
    desired_cycle = desired_cycle.join(split_index_to_cols(desired_cycle))
    unique_gene_inhibition = desired_cycle.Gene_inhibition.unique()

    if len(unique_gene_inhibition) >1 or '_' not in unique_gene_inhibition[0]:
        desired_cycle = desired_cycle.set_index('Gene_inhibition')[['cycle_max_gr', 'bool_growing', 'growth_phase','growth_phase_length', 'Species','culture','end_cycle']]
    else:
        desired_cycle.index = ['_'.join([x[1],x[3]]) for x in desired_cycle.index.str.split('_')]
        desired_cycle.Gene_inhibition = desired_cycle.index

    return desired_cycle

def convert_col_to_gene_inhibition(biomass_column):
    if len(biomass_column.split('.')) <=2:
        _, gcomb, _ = biomass_column.split('_')
        return gcomb
    _, gcomb, _, lv_pairs = biomass_column.split('_')
    return gcomb + '_' + lv_pairs

def get_end_BM(biomass_df):    
    """retrieve biomass at end cycle"""
    def get_species_frac_binned(end_biomass): # derive species relative abundance in media
        stable_species_frac = end_biomass.reset_index('Species', drop=True)['BM_consortia_frac']
        # stable_species_frac = end_biomass.query('Species=="E0"').reset_index('Species', drop=True)['BM_consortia_frac'] # whyv need query E0

        labels = ['S', 'slight E', 'E']
        bins = [0, 0.5, 0.7, 1]
        stable_species_frac_binned = pd.cut(stable_species_frac, bins=bins, labels=labels)
        return stable_species_frac_binned
    
    def add_to_end_biomass(end_biomass): # additional columns counting species relative population 
        def get_ratio_and_std_col(model_id):
            temp_df = separate_Species_df(end_biomass, model_id, inc_Species=True)
            temp_df['BM_consortia_frac'] = temp_df.Biomass/temp_df.Total_BM
            temp_df['standardized_BM'] = temp_df.Biomass/temp_df.loc['Normal', 'Biomass']
            return temp_df
        
        def separate_Species_df(df, model_id, inc_Species = False):
            """separate species from df"""
            def get_species_loc(model_id):
                return list(temp_df.Species == model_id)
            
            if not isinstance(model_id, str):
                model_id = model_id.id
            df = df.loc[:,~df.columns.str.contains('Ndiff|ESdiff')]
            temp_df = pd.concat([df['Species'],df.select_dtypes(include = ['float'])], axis=1)
            result_df = temp_df.loc[get_species_loc(model_id)]
            return result_df if inc_Species else result_df.drop('Species',axis=1)
        return pd.concat([get_ratio_and_std_col('E0'), get_ratio_and_std_col('S0')]).set_index('Species', append=True).drop(['Total_BM', 'Biomass'], axis=1)

    biomass = pd.DataFrame(biomass_df.loc[:, [col for col in biomass_df.columns if 'coculture' in col]].iloc[-1])
    biomass.columns = ['Biomass']
    biomass['Species'] = list(pd.Series(biomass.index).apply(lambda x: str(x).split('_')[0]))
    biomass['Gene_inhibition'] = list(pd.Series(biomass.index).apply(convert_col_to_gene_inhibition))
    biomass = biomass.set_index(['Gene_inhibition'])
    biomass['Total_BM'] = biomass.groupby('Gene_inhibition').Biomass.sum()
    end_biomass = set_GI_SP_as_MI(biomass).join(set_GI_SP_as_MI(add_to_end_biomass(biomass)))
    binned_col = get_species_frac_binned(end_biomass)
    # return end_biomass, binned_col

    # end_biomass = end_biomass.merge(binned_col, left_index=True, right_index=True, suffixes=('', '_binned'))
    # end_biomass['BM_consortia_frac_binned'] = end_biomass['BM_consortia_frac_binned'].cat.add_categories(['No growth'])
    end_biomass['BM_consortia_frac_binned'] = list(binned_col)
    end_biomass.loc[end_biomass.Biomass<1.01e-8,'BM_consortia_frac_binned'] = 'No growth'
    return end_biomass
